import re


def first_or_none(values=None):
    if values and type(values) is list and len(values) > 0:
        return values[0]
    return None


def cast_bool(f=None):
    return True if f.lower() in ["true", "1", "yes"] else False


def strip_supressed(f=None):
    return None if f.lower() in ["suppressed"] else f


def strip_schema(f=None):
    return f


def strip_spaces(f=None):
    return f.replace(" ", "")


def capitalize(f=None):
    return f.capitalize()



import enum


class DispatchType(enum.Enum):
    GENERATOR = 1
    LOAD = 2


def parse_dispatch_type(dispatch_string: str):
    """
        Converts dispatch type string into an enum constant

    """
    if not dispatch_string:
        return None

    dispatch_string = dispatch_string.lower().strip()

    if dispatch_string == "load":
        return DispatchType.LOAD

    if dispatch_string == "generating":
        return DispatchType.GENERATOR

    if dispatch_string == "generator":
        return DispatchType.GENERATOR

    raise Exception("Unknown dispatch type: {}".format(dispatch_string))



import pandas as pd
from dms_daemon import orm

NEMWEB_META = orm.engine_meta
NEMWEB_DERIVED = orm.engine_derived

def load_gen_meta():
	#load "full register"
	sql = 	"SELECT R.* "\
			"FROM nemweb_meta.REGISTER R "\
			"INNER JOIN "\
				"(SELECT DUID, MAX(DATE) AS LATEST_DATE "\
				 "FROM REGISTER "\
				"GROUP BY DUID) AS D "\
			"ON R.DATE = D.LATEST_DATE and D.DUID = R.DUID"
	return pd.read_sql(sql, con=NEMWEB_META, index_col=None)
	
def duid_keys():
    sql = "SELECT ID, DUID FROM DUID"
    results = NEMWEB_META.execute(sql)
    return dict(results.fetchall())

def category_map():
    sql =     "SELECT NAME,ID FROM CATEGORY"
    results = NEMWEB_DERIVED.execute(sql)
    return dict(results.fetchall())

def region_map():
    sql =   "SELECT R.DUID, R.REGIONID "\
            "FROM REGISTER R "\
            "INNER JOIN LAST_REGISTERED_DATE LRD "\
                "ON LRD.LAST_DATE = R.DATE "\
            "AND LRD.DUID = R.DUID"
    results = NEMWEB_META.execute(sql)
    rdict = dict(results.fetchall())
    manual_mapping(rdict)
    return rdict

def manual_mapping(rdict):
    for duid, regionid in {724: 3,  #ancilliary Service SA (ASSENC1)
                           732: 1,  #ancialliary Service NSW (ASNACTW1)
                           747: 3,
                           753: 5,  #?Not sure - VSSEL1V1
                           763: 5,
                           767: 1,
                           }.items():
        if duid not in rdict:
            rdict[duid] = regionid
        else:
            print ("DUID exists ({0})".format(duid))

REGION_MAP = region_map()
DUID_KEYS = duid_keys()
CATEGORY_MAP = category_map()

#unused
gen_meta = load_gen_meta()




import click
from scrapy.utils.python import garbage_collect

from opennem.diff.facility_diff import run_diff
from opennem.importer.opennem import run_opennem_import
from opennem.utils import logging

logger = logging.getLogger("opennem.cli")
logger.setLevel(logging.DEBUG)


@click.group()
def main():
    pass


@click.command()
def crawl():
    logger.info("crawl")


@click.command()
def diff():
    run_diff()


@click.command()
def cmd_import():
    run_opennem_import()


main.add_command(crawl)
main.add_command(diff)
main.add_command(cmd_import, name="import")

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        logger.error("User interrupted")
    except Exception as e:
        logger.error(e)
    finally:
        garbage_collect()



import csv
import json
import os
from typing import List, Optional

DATA_PATH = os.path.join(os.path.dirname(__file__), "data")


class FileNotFound(Exception):
    pass


def load_data_json(fixture_name):
    fixture_path = os.path.join(DATA_PATH, fixture_name)

    if not os.path.isfile(fixture_path):
        raise FileNotFound("Not a file: {}".format(fixture_path))

    fixture = None

    with open(fixture_path) as fh:
        fixture = json.load(fh)

    return fixture


def load_data_csv(filename: str) -> Optional[List[dict]]:
    """
        Load a CSV file

        @TODO use libmagic to determine encoding types since Excel saves
        can be funky
    """

    data_filepath = os.path.join(DATA_PATH, filename)

    if not os.path.isfile(data_filepath):
        raise FileNotFound("Not a file: {}".format(data_filepath))

    records = []

    # leave the encoding in place now todo is to determine it
    with open(data_filepath, encoding="utf-8-sig") as fh:
        csvreader = csv.DictReader(fh)
        records = [entry for entry in csvreader]

    return records



import os

from opennem.core import load_data_csv
from opennem.db.models.opennem import Facility, Station
from opennem.utils import logging

logger = logging.getLogger("opennem.importer")

RECORD_MODEL_MAP = {
    "STATION": Station,
    "FACILITY": Facility,
}


def run_opennem_import():
    """
        Reads the OpenNEM data source

    """

    opennem_records = load_data_csv("opennem.csv")

    for rec in opennem_records:
        logger.debug(rec)

        if not "record_type" in rec:
            raise Exception("Invalid CSV: No record_type")

        record_type = rec["record_type"]

        if not record_type in RECORD_MODEL_MAP:
            raise Exception(
                "Invalid record type: {} is not a valid record type".format(
                    record_type
                )
            )

        record_model = RECORD_MODEL_MAP[record_type]



"""

Google Places API

Methods to use and access the Google Places API

    * encode place names to geolocation using Google Places API
    * find a place id
    * autocomplete

"""
import logging
import os
import sys
import time
from datetime import timedelta

import requests
import requests_cache

from opennem.settings import GOOGLE_PLACES_API_KEY, REQUESTS_CACHE_PATH

logging.basicConfig(level=logging.INFO)


requests_cache.install_cache(
    REQUESTS_CACHE_PATH, expire_after=timedelta(days=60)
)

BACKOFF_TIME = 30
QUERY_LIMIT = 0


def google_geocode(
    query, region=None, api_key=None, return_full_response=False
):

    GOOGLE_PLACES_URL = "https://maps.googleapis.com/maps/api/geocode/json"

    query_params = {}

    query_components = {}

    url_params = {
        "key": GOOGLE_PLACES_API_KEY,
    }

    if region:
        query_params["region"] = region

    if "postal_code" in query:
        query_components["postal_code"] = query["postal_code"]

    if "country" in query:
        query_params["region"] = query["country"]
        query_components["country"] = query["country"]
        query.pop("country", None)

    url_params["components"] = "|".join(
        f"{k}:{v}" for k, v in query_components.items()
    )

    results = requests.get(GOOGLE_PLACES_URL, params=url_params).json()

    logging.debug(results)

    if not "status" in results:
        raise Exception("Invalid response")

    if results["status"] == "REQUEST_DENIED":
        raise Exception("API Key or other request denied error")

    if results["status"] == "OVER_QUERY_LIMIT":
        logging.warn("Hit Query Limit! Backing off for a bit.")
        time.sleep(5)  # sleep for 30 minutes
        return google_geocode(query, region, api_key, return_full_response)

    if results["status"] == "ZERO_RESULTS":
        return None

    if not results["status"] == "OK":
        logging.error("Error results for %s", query)
        raise Exception("No results: {}".format(results["status"]))

    if len(results["results"]) == 0:
        logging.error("No results for %s", query)
        return None

    cand = results["results"][0]
    return cand


def lookup_placeid(place_id, api_key=None, retry=0):

    GOOGLE_PLACES_URL = (
        "https://maps.googleapis.com/maps/api/place/details/json"
    )

    url_params = {"key": GOOGLE_PLACES_API_KEY, "place_id": place_id}

    results = requests.get(GOOGLE_PLACES_URL, params=url_params).json()

    if not "status" in results:
        raise Exception("Invalid response")

    if results["status"] == "REQUEST_DENIED":
        raise Exception("API Key or other request denied error")

    if results["status"] == "OVER_QUERY_LIMIT":
        logging.warn("Hit Query Limit! Backing off for a bit.")
        time.sleep(5)  # sleep for 30 minutes
        lookup_placeid(place_id)

    if results["status"] == "ZERO_RESULTS":
        return None

    if not results["status"] == "OK":
        logging.error("Error results for %s", place_id)
        raise Exception("No results: {}".format(results["status"]))

    if not "result" in results:
        logging.error("No result for %s", place_id)
        raise Exception("No result: {}".format(place_id))

    return results["result"]


def place_autocomplete(
    query, region=None, api_key=None, return_full_response=False
):

    GOOGLE_PLACES_URL = (
        "https://maps.googleapis.com/maps/api/place/autocomplete/json"
    )

    query_params = {}

    query_components = {}

    url_params = {
        "key": GOOGLE_PLACES_API_KEY,
        "input": query,
    }

    if region:
        query_params["region"] = region

    if "postal_code" in query:
        query_components["postal_code"] = query["postal_code"]

    if "country" in query:
        query_params["region"] = query["country"]
        query_components["country"] = query["country"]
        query.pop("country", None)

    url_params["components"] = "|".join(
        f"{k}:{v}" for k, v in query_components.items()
    )

    results = requests.get(GOOGLE_PLACES_URL, params=url_params).json()

    logging.debug(results)

    if not "status" in results:
        raise Exception("Invalid response")

    if results["status"] == "REQUEST_DENIED":
        raise Exception("API Key or other request denied error")

    if results["status"] == "OVER_QUERY_LIMIT":
        logging.warn("Hit Query Limit! Backing off for a bit.")
        time.sleep(5)  # sleep for 30 minutes
        return google_geocode(query, region, api_key, return_full_response)

    if results["status"] == "ZERO_RESULTS":
        return None

    if not results["status"] == "OK":
        logging.error("Error results for %s", query)
        raise Exception("No results: {}".format(results["status"]))

    if len(results["predictions"]) == 0:
        logging.error("No results for %s", query)
        return None

    cand = results["predictions"]
    return cand


def place_search(query, api_key=None, return_full_response=False):

    GOOGLE_PLACES_URL = (
        "https://maps.googleapis.com/maps/api/place/findplacefromtext/json"
    )

    query_params = {}

    query_components = {}

    url_params = {
        "key": GOOGLE_PLACES_API_KEY,
        "input": query,
        "inputtype": "textquery",
        "language": "en-AU",
        # "fields": "basic"
        "fields": "name,geometry,place_id",
    }

    results = requests.get(GOOGLE_PLACES_URL, params=url_params).json()

    logging.debug(results)

    if not "status" in results:
        raise Exception("Invalid response")

    if results["status"] == "REQUEST_DENIED":
        raise Exception("API Key or other request denied error")

    if results["status"] == "OVER_QUERY_LIMIT":
        logging.warn("Hit Query Limit! Backing off for a bit.")
        time.sleep(5)  # sleep for 30 minutes
        return place_search(query)

    if results["status"] == "ZERO_RESULTS":
        return None

    if not results["status"] == "OK":
        logging.error("Error results for %s", query)
        raise Exception("No results: {}".format(results["status"]))

    if len(results["candidates"]) == 0:
        logging.error("No results for %s", query)
        return None

    cand = results["candidates"]
    return cand



import json
import os

DATA_PATH = os.path.join(os.path.dirname(__file__), "data")


def load_single_units(fixture_name):
    fixture_path = os.path.join(DATA_PATH, fixture_name)

    if not os.path.isfile(fixture_path):
        raise Exception("Not a file: {}".format(fixture_path))

    fixture = None

    with open(fixture_path) as fh:
        fixture = json.load(fh)

    return fixture


SINGLE_UNITS = load_single_units("single_units.json")


def facility_unit_numbers_are_single(facility_duid):
    """
        These units have unit numbers that are single units rather than ranges.

        ex. GT 1-2 means unit of alias GT1 and id 2 rather than alias GT and range 1-2

    """
    if not type(SINGLE_UNITS) is list:
        raise Exception("Error loading facility station map: not a list")

    return facility_duid in SINGLE_UNITS



import json
import os

DATA_PATH = os.path.join(os.path.dirname(__file__), "data")


def load_facility_stations(fixture_name):
    fixture_path = os.path.join(DATA_PATH, fixture_name)

    if not os.path.isfile(fixture_path):
        raise Exception("Not a file: {}".format(fixture_path))

    fixture = None

    with open(fixture_path) as fh:
        fixture = json.load(fh)

    return fixture


FACILITY_STATIONS = load_facility_stations("facility_stations.json")


def facility_station_join_by_name(station_name):
    """
        This is a bit of a hack where we join facilities to stations by name only
        where there is no duid or other data we can join on.

        Used a lot in updating AEMO data.

        The list of station names to do this based on is stored in data/facility_stations.json

    """
    if not type(FACILITY_STATIONS) is list:
        raise Exception("Error loading facility station map: not a list")

    return station_name not in FACILITY_STATIONS



import decimal
import json
from datetime import datetime

from geojson import GeoJSONEncoder


class OpenNEMJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)
        if isinstance(o, datetime):
            return o.isoformat()
        return super(OpenNEMGeoJSONEncoder, self).default(o)


class OpenNEMGeoJSONEncoder(GeoJSONEncoder, OpenNEMJSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)
        if isinstance(o, datetime):
            return o.isoformat()
        return super(OpenNEMGeoJSONEncoder, self).default(o)



import json
import logging
from datetime import datetime

from scrapy.exceptions import DropItem
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy.sql import text

from opennem.core.facilitystations import facility_station_join_by_name
from opennem.core.facilitystatus import lookup_facility_status
from opennem.core.fueltechs import lookup_fueltech
from opennem.core.normalizers import (
    clean_capacity,
    name_normalizer,
    normalize_duid,
    participant_name_filter,
    station_name_cleaner,
)
from opennem.db.models.opennem import (
    Facility,
    FacilityStatus,
    Participant,
    Station,
)
from opennem.pipelines import DatabaseStoreBase
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class NemStoreGI(DatabaseStoreBase):
    def parse_interval(self, date_str):
        return datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        if item["FuelType"] in ["Natural Gas Pipeline"]:
            return {}

        if (
            item["UnitStatus"]
            in [
                "Publically Announced",
                "Upgrade",
                "Emerging",
                "Expansion",
                "Maturing",
            ]
            or not item["UnitStatus"]
        ):
            return {}

        participant_name = name_normalizer(item["Owner"])

        # Funky case of Todae solar where they put their name in the

        participant = (
            s.query(Participant)
            .filter(Participant.name == participant_name)
            .one_or_none()
        )

        if not participant:
            participant = Participant(
                name=participant_name,
                name_clean=participant_name_filter(participant_name),
            )

            s.add(participant)
            s.commit()
            logger.info(
                "GI: Added new partipant to NEM database: {}".format(
                    participant_name
                )
            )

        facility = None
        facility_station = None

        facility_name = name_normalizer(item["Name"])
        facility_name_clean = station_name_cleaner(item["Name"])
        duid = normalize_duid(item["DUID"])
        facility_status = lookup_facility_status(item["UnitStatus"])
        facility_region = item["Region"]
        facility_fueltech = (
            lookup_fueltech(item["FuelType"], techtype=item["TechType"])
            if ("FuelType" in item and item["FuelType"])
            else None
        )

        if facility_station_join_by_name(facility_name_clean):
            facility_station = (
                s.query(Station)
                .filter(Station.name_clean == facility_name_clean)
                .first()
            )

        if duid:
            try:
                facility = (
                    s.query(Facility)
                    .filter(Facility.code == duid)
                    .filter(Facility.region == facility_region)
                    # .filter(Facility.nameplate_capacity != None)
                    .one_or_none()
                )
            except MultipleResultsFound:
                logger.warn(
                    "Multiple results found for duid : {}".format(duid)
                )

            if facility:
                if not facility.station:
                    raise Exception(
                        "GI: Existing facility {} {} with no station .. unpossible.".format(
                            facility.id, facility.name
                        )
                    )

                if facility.station and not facility_station:
                    facility_station = facility.station

                logger.info(
                    "GI: Found facility by DUID: {} {} {}".format(
                        facility.id, facility.name, facility_station.id
                    )
                )

        if not duid:
            facility_lookup = (
                s.query(Facility)
                .filter(Facility.region == facility_region)
                .filter(Facility.name_clean == facility_name_clean)
                # .filter(Facility.fueltech_id == facility_fueltech)
                .filter(Facility.code == None)
                .first()
            )

            if facility_lookup:
                if not facility_lookup.station:
                    raise Exception(
                        "GI: Existing facility {} {} with no station .. unpossible.".format(
                            facility.id, facility.name
                        )
                    )

                if facility_lookup.station and not facility_station:
                    facility_station = facility_lookup.station

                # logger.info(
                #     "GI: Found facility by no duid name fueltech and region : {} {} {}".format(
                #         facility.id, facility.name, facility_station.id
                #     )
                # )

        # Done trying to find existing

        created_station = False
        created_facility = False

        if not facility_station:
            facility_station = Station(
                name=item["Name"],
                name_clean=facility_name_clean,
                nem_region=facility_region,
                created_by="pipeline.nem.facility.gi",
            )

            s.add(facility_station)

            created_station = True

        if not facility:
            facility = Facility(created_by="pipeline.nem.facility.gi")
            facility.station = facility_station

            created_facility = True

        if duid:
            facility.code = duid

        facility.participant = participant

        if not facility.region:
            facility.region = facility_region

        if not facility.name:
            facility.name = name_normalizer(item["Name"])

        if not facility.name_clean:
            facility.name_clean = station_name_cleaner(item["Name"])

        # @TODO parse units into ints
        facility.unit_number = item["Units"]

        if not facility.fueltech_id and facility_fueltech:
            facility.fueltech_id = facility_fueltech

        if facility.fueltech_id is None:
            logger.error("Could not find fueltech for: {}".format(item))

        if not facility.nameplate_capacity:
            facility.nameplate_capacity = clean_capacity(
                item["UpperCapacity"] or item["NameCapacity"]
            )

        facility.status_id = facility_status

        if facility_station and not facility.station:
            facility.station = facility_station

        if facility.status_id is None:
            raise Exception(
                "GI: Failed to map status ({}) on row: {}".format(
                    facility.status_id, item
                )
            )

        if created_station:
            logger.info(
                "GI: {} station with name {} and id {}".format(
                    "Created" if created_station else "Updated",
                    facility_station.name_clean,
                    facility_station.id,
                )
            )

        if created_facility:
            logger.info(
                "GI: {} facility with name {} and duid {} and id {}".format(
                    "Created" if created_facility else "Updated",
                    facility.name_clean,
                    facility.code,
                    facility.id,
                )
            )

        try:
            s.add(facility)
            s.commit()
        except Exception as e:
            logger.error(e)
            raise e
        finally:
            s.close()

        return {
            "name": name_normalizer(item["Name"]),
            "status": item["UnitStatus"],
            "status_id": lookup_facility_status(item["UnitStatus"]),
        }


class NemStoreMMS(DatabaseStoreBase):
    def parse_interval(self, date_str):
        return datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        print(item)



import json

from opennem.controllers.stations import get_stations
from opennem.exporter.csv import stations_csv_records
from opennem.exporter.encoders import OpenNEMJSONEncoder


def stations_json_serialize():

    json_records = stations_csv_records()

    json_serialized = json.dumps(
        json_records, cls=OpenNEMJSONEncoder, indent=4
    )

    return json_serialized



from typing import Optional

from pydantic import BaseModel, ValidationError, validator

from opennem.core.normalizers import (
    clean_capacity,
    is_number,
    normalize_duid,
    normalize_string,
    participant_name_filter,
    station_name_cleaner,
)

from .opennem import OpennemBaseModel

__all__ = ["OpennemStationSubmission", "OpennemStation"]


class OpennemStationSubmission(BaseModel):
    name: str
    network_id: str


class OpennemStation(OpennemBaseModel):
    code: Optional[str]
    name: str

    @classmethod
    @validator("code")
    def code_clean(cls, v):
        return normalize_duid(v)

    @classmethod
    @validator("network_name")
    def network_name_isalpha(cls, v):
        assert type(v) is str, "must be string"
        # assert v.isalpha(), "must be alpha: {}".format(v)

        return v

    @classmethod
    @validator("name")
    def name_clean(cls, v):
        # todo auto-set name based on network_name
        name = participant_name_filter(v)

        return name



from opennem.exporter.aws import write_to_s3
from opennem.exporter.csv import stations_csv_serialize
from opennem.exporter.geojson import stations_geojson_serialize
from opennem.exporter.local import write_to_local
from opennem.exporter.onjson import stations_json_serialize
from opennem.utils import logging

logger = logging.getLogger("opennem.exporter")


def stations_geojson_to_s3():
    stations_geojson = stations_geojson_serialize()

    facilities_path = "geo/au_facilities.json"

    file_length = write_to_s3("geo/au_facilities.json", stations_geojson)


def stations_geojson_to_local():
    stations_geojson = stations_geojson_serialize()

    write_to_local("stations.geojson", stations_geojson)


def stations_csv_to_local():
    stations_csv = stations_csv_serialize()

    write_to_local("stations.csv", stations_csv)


def stations_json_to_local():
    stations_json = stations_json_serialize()

    write_to_local("stations.json", stations_json)


if __name__ == "__main__":
    stations_geojson_to_s3()
    stations_geojson_to_local()
    stations_csv_to_local()
    stations_json_to_local()



import decimal
import json
import logging
from datetime import date, datetime
from decimal import Decimal
from pprint import pprint
from typing import List, Optional

from geojson import Feature, FeatureCollection, Point, dumps
from sqlalchemy import func

from opennem.db.models.opennem import (
    Facility,
    FacilityStatus,
    FuelTech,
    Network,
    Station,
    metadata,
)
from opennem.schema.opennem import StationSubmission

logger = logging.getLogger(__name__)


def get_stations(
    session, name: str = None, limit: Optional[int] = None, page: int = 1
) -> List[Station]:
    """
        API controller that gets all stations sorted and joined

    """

    stations = (
        session.query(Station)
        .join(Facility)
        .join(Facility.fueltech)
        .filter(Facility.fueltech_id.isnot(None))
        .filter(Facility.status_id.isnot(None))
    )

    if name:
        stations = stations.filter(Station.name.like("%{}%".format(name)))

    stations = stations.order_by(
        Facility.network_region,
        Station.name,
        Facility.network_code,
        Facility.code,
    )

    stations = stations.all()

    logger.info("Got {} stations".format(len(stations)))

    return stations


def get_station(session, station_id: str) -> Station:
    """
        Returns a single station by id

    """
    station = session.query(Station).get(station_id)

    return station


def create_station(session, station: StationSubmission) -> Station:
    """
        Create a station

    """
    station_record = Station(name=station.name, network_id=station.network_id)

    session.add(station_record)
    session.commit()

    return station_record



from opennem.spiders.bom.bom_base import BomJSONObservationSpider


class BomSydneySpider(BomJSONObservationSpider):
    name = "bom.capitals.sydney"
    start_urls = ["http://reg.bom.gov.au/fwo/IDN60901/IDN60901.94767.json"]
    observatory_id = "066037"


class BomMelbourneSpider(BomJSONObservationSpider):
    name = "bom.capitals.melbourne"
    start_urls = ["http://reg.bom.gov.au/fwo/IDV60901/IDV60901.94866.json"]
    observatory_id = "086282"


class BomBrisbaneSpider(BomJSONObservationSpider):
    name = "bom.capitals.brisbane"
    start_urls = ["http://reg.bom.gov.au/fwo/IDQ60901/IDQ60901.94576.json"]
    observatory_id = "040913"


class BomPerthSpider(BomJSONObservationSpider):
    name = "bom.capitals.perth"
    start_urls = ["http://reg.bom.gov.au/fwo/IDW60901/IDW60901.94610.json"]
    observatory_id = "009021"



import scrapy

from opennem.pipelines.nem import (
    ExtractCSV,
    ReadStringHandle,
    TableRecordSplitter,
    UnzipSingleFilePipeline,
)
from opennem.pipelines.nem.mms import (
    NemStoreMMSDudetail,
    NemStoreMMSDudetailSummary,
    NemStoreMMSParticipant,
    NemStoreMMSStatdualloc,
    NemStoreMMSStations,
    NemStoreMMSStationStatus,
)
from opennem.pipelines.nem.mms_dudetailsummary import (
    AEMOMMSDudetailSummaryGrouper,
)
from opennem.spiders.dirlisting import DirlistingSpider
from opennem.spiders.nem import NemSingleMMSSpider


class NemArchiveMMSSpider(DirlistingSpider):
    name = "au.nem.archive.mms"

    start_urls = [
        "http://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM/2020/MMSDM_2020_06/MMSDM_Historical_Data_SQLLoader/DATA/"
    ]


class NemMMSStationSpider(NemSingleMMSSpider):
    name = "au.nem.mms.stations"

    pipelines_extra = set([NemStoreMMSStations])
    url = "http://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM/2020/MMSDM_2020_06/MMSDM_Historical_Data_SQLLoader/DATA/PUBLIC_DVD_STATION_202006010000.zip"


class NemMMSStationStatusSpider(NemSingleMMSSpider):
    name = "au.nem.mms.stations_status"

    pipelines_extra = set([NemStoreMMSStationStatus])
    url = "http://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM/2020/MMSDM_2020_06/MMSDM_Historical_Data_SQLLoader/DATA/PUBLIC_DVD_STATIONOPERATINGSTATUS_202006010000.zip"


class NemMMSStationStatusSpider(NemSingleMMSSpider):
    name = "au.nem.mms.participant"

    pipelines_extra = set([NemStoreMMSParticipant])
    url = "http://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM/2020/MMSDM_2020_06/MMSDM_Historical_Data_SQLLoader/DATA/PUBLIC_DVD_PARTICIPANT_202006010000.zip"


class NemMMSDudeTailSpider(NemSingleMMSSpider):
    name = "au.nem.mms.dudetail"

    pipelines_extra = set([NemStoreMMSDudetail,])
    url = "http://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM/2020/MMSDM_2020_06/MMSDM_Historical_Data_SQLLoader/DATA/PUBLIC_DVD_DUDETAIL_202006010000.zip"


class NemMMSDudeTailSummarySpider(NemSingleMMSSpider):
    name = "au.nem.mms.dudetail_summary"

    # Override the base pipelines since we're using a grouper
    # @TODO make this generic
    pipelines = set(
        [
            UnzipSingleFilePipeline,
            ReadStringHandle,
            ExtractCSV,
            AEMOMMSDudetailSummaryGrouper,
        ]
    )

    pipelines_extra = set([NemStoreMMSDudetailSummary,])
    url = "http://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM/2020/MMSDM_2020_06/MMSDM_Historical_Data_SQLLoader/DATA/PUBLIC_DVD_DUDETAILSUMMARY_202006010000.zip"


class NemMMSStaduAlloc(NemSingleMMSSpider):
    name = "au.nem.mms.statdualloc"

    pipelines_extra = set([NemStoreMMSStatdualloc,])
    url = "http://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM/2020/MMSDM_2020_06/MMSDM_Historical_Data_SQLLoader/DATA/PUBLIC_DVD_STADUALLOC_202006010000.zip"



import json
import logging
from datetime import datetime
from itertools import groupby

from scrapy.exceptions import DropItem
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy.sql import text

from opennem.core.facilitystations import facility_station_join_by_name
from opennem.core.facilitystatus import lookup_facility_status
from opennem.core.fueltechs import lookup_fueltech
from opennem.core.normalizers import (
    clean_capacity,
    name_normalizer,
    normalize_aemo_region,
    normalize_duid,
    participant_name_filter,
    station_name_cleaner,
)
from opennem.core.station_duid_map import (
    facility_has_station_remap,
    facility_map_station,
)
from opennem.core.unit_codes import get_unit_code
from opennem.core.unit_parser import parse_unit_duid
from opennem.db.models.opennem import (
    Facility,
    FacilityStatus,
    Participant,
    Station,
)
from opennem.exporter.encoders import OpenNEMJSONEncoder
from opennem.pipelines import DatabaseStoreBase
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


FACILITY_INVALID_STATUS = [
    "Publically Announced",
    "Upgrade",
    "Emerging",
    "Expansion",
    "Maturing",
]


def get_station_record_from_facilities(units: list):
    if not type(units) is list or len(units) < 1:
        raise Exception("Passed units list with no units ..")

    if len(units) == 1:
        return units[0]

    for u in units:
        cap = clean_capacity(u["NameCapacity"])
        duid = normalize_duid(u["duid"])

        if cap:
            return u

    return units[0]


def record_get_station_name(facilities) -> str:
    station_name = facilities[0]["station_name"]

    return station_name


def has_unique_duid(units: list) -> bool:
    duids = set([i["duid"] for i in units])
    return len(duids) == len(units)


def get_unique_duid(units: list) -> str:
    if not type(units) is list or len(units) < 1:
        return None

    first_record = units[0]

    return (
        normalize_duid(first_record["duid"])
        if "duid" in first_record
        else None
    )


def get_unique_reqion(units: list) -> str:
    if not type(units) is list or len(units) < 1:
        return None

    first_record = units[0]

    return first_record["Region"].strip()


class GeneralInformationGrouperPipeline(object):
    """
        This is the first-pass pipeline of the AEMO general information list

        It groups by clean station name

    """

    @check_spider_pipeline
    def process_item(self, item, spider=None):
        if not "records" in item:
            raise Exception("No records found in item pipeline failed")

        generators = item["records"]

        generators = list(
            filter(lambda x: x["station_name"] != None, generators)
        )

        # sort by name

        generators_grouped = {}

        for k, v in groupby(
            generators, key=lambda v: (v["SurveyID"], v["station_name"])
        ):
            key = k[0]
            if not key in generators_grouped:
                generators_grouped[key] = []

            generators_grouped[key] += list(v)

        # generators_grouped = list(generators_grouped.items()

        return {"generators": generators_grouped}


class GeneralInformationStoragePipeline(DatabaseStoreBase):
    """
        Stores the grouped General Information items from AEMO

    """

    def parse_interval(self, date_str):
        return datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")

    def process_participants(self, participants):
        s = self.session()

        for participant in participants:
            participant_name = name_normalizer(participant["Owner"])

            # Funky case of Todae solar where they put their name in the

            participant = (
                s.query(Participant)
                .filter(Participant.name == participant_name)
                .one_or_none()
            )

            if not participant:
                participant = Participant(
                    name=participant_name,
                    name_clean=participant_name_filter(participant_name),
                )

                s.add(participant)
                # s.commit()
                logger.info(
                    "GI: Added new partipant to NEM database: {}".format(
                        participant_name
                    )
                )

    def process_facilities(self, records):
        s = self.session()

        all_duids = [
            i[0]
            for i in s.query(Facility.network_code)
            .filter(Facility.network_code != None)
            .all()
        ]

        for _, facility_records in records.items():
            facility_index = 1
            facility_station = None
            created_station = False

            station_network_name = record_get_station_name(facility_records)
            station_name = station_name_cleaner(station_network_name)

            duid_unique = has_unique_duid(facility_records)
            facility_count = len(facility_records)

            # Step 1. Find the station
            # First by duid if it's unique
            duid = get_unique_duid(facility_records)

            # all GI records should have a region
            station_network_region = get_unique_reqion(facility_records)

            # This is the most suitable unit record to use for the station
            # see helper above
            facility_station_record = get_station_record_from_facilities(
                facility_records
            )

            if duid and duid_unique and facility_count == 1:

                facility_lookup = None

                try:
                    facility_lookup = (
                        s.query(Facility)
                        .filter(Facility.network_code == duid)
                        .filter(Facility.network_code != None)
                        .filter(
                            Facility.network_region == station_network_region
                        )
                        .one_or_none()
                    )
                except MultipleResultsFound:
                    logger.error(
                        "Found multiple duid for station with code {}".format(
                            duid
                        )
                    )
                    continue

                if facility_lookup and facility_lookup.station:
                    facility_station = facility_lookup.station

            if (
                duid
                and (duid_unique and facility_count > 1)
                or not duid_unique
            ):

                facility_lookup = (
                    s.query(Facility)
                    .filter(Facility.network_code == duid)
                    .filter(Facility.network_region == station_network_region)
                    .filter(Facility.network_code != None)
                    .first()
                )

                if facility_lookup and facility_lookup.station:
                    facility_station = facility_lookup.station

            if not facility_station and facility_station_join_by_name(
                station_name
            ):
                try:
                    facility_station = (
                        s.query(Station)
                        .filter(Station.name == station_name)
                        .one_or_none()
                    )
                except MultipleResultsFound:
                    logger.warning(
                        "Multiple results found for station name : {}".format(
                            station_name
                        )
                    )
                    facility_station = None

            # If we have a station name, and no duid, and it's ok to join by name
            # then find the station (make sure to region lock)
            if (
                station_name
                and not duid
                and not facility_station
                and facility_station_join_by_name(station_name)
            ):
                facility = (
                    s.query(Facility)
                    .join(Facility.station)
                    .filter(Facility.network_region == station_network_region)
                    .filter(Station.name == station_name)
                    .first()
                )

                if facility:
                    facility_station = facility.station

            # Create one as it doesn't exist
            if not facility_station:
                facility_station = Station(
                    name=station_name,
                    network_name=name_normalizer(
                        facility_station_record["station_name"]
                    ),
                    network_id="NEM",
                    created_by="pipeline.aemo.general_information",
                )

                s.add(facility_station)
                s.commit()

                created_station = True
            else:
                facility_station.updated_by = (
                    "pipeline.aemo.general_information"
                )

            for facility_record in facility_records:
                # skip pipelines
                if facility_record["FuelType"] in ["Natural Gas Pipeline"]:
                    continue

                # skip these statuses too
                if facility_record["UnitStatus"] in FACILITY_INVALID_STATUS:
                    continue

                facility = None
                created_facility = False

                facility_network_name = name_normalizer(
                    facility_record["station_name"]
                )
                facility_name = station_name_cleaner(
                    facility_record["station_name"]
                )
                duid = normalize_duid(facility_record["duid"])
                reg_cap = clean_capacity(facility_record["NameCapacity"])

                units_num = facility_record["Units"] or 1
                unit_id = facility_index + (units_num - 1)

                unit = parse_unit_duid(unit_id, duid)
                unit_size = clean_capacity(facility_record["unit_capacity"])
                unit_code = get_unit_code(
                    unit, duid, facility_record["station_name"]
                )

                facility_comissioned = facility_record["SurveyEffective"]
                facility_comissioned_dt = None

                if type(facility_comissioned) is datetime:
                    facility_comissioned_dt = facility_comissioned

                try:
                    if type(facility_comissioned) is str:
                        facility_comissioned_dt = datetime.strptime(
                            facility_comissioned, "%d/%m/%y"
                        )
                except ValueError:
                    logger.error(
                        "Error parsing date: {}".format(facility_comissioned)
                    )

                facility_status = lookup_facility_status(
                    facility_record["UnitStatus"]
                )
                facility_network_region = normalize_aemo_region(
                    facility_record["Region"]
                )
                facility_fueltech = (
                    lookup_fueltech(
                        facility_record["FuelType"],
                        techtype=facility_record["TechType"],
                    )
                    if (
                        "FuelType" in facility_record
                        and facility_record["FuelType"]
                    )
                    else None
                )

                if not facility_fueltech:
                    logger.error(
                        "Error looking up fueltech: {} {} ".format(
                            facility_record["FuelType"],
                            facility_record["TechType"],
                        )
                    )

                # check if we have it by ocode first
                facility = (
                    s.query(Facility)
                    .filter(Facility.code == unit_code)
                    .one_or_none()
                )

                if not facility and duid:
                    try:
                        facility = (
                            s.query(Facility)
                            .filter(Facility.network_code == duid)
                            .filter(
                                Facility.network_region
                                == facility_network_region
                            )
                            # .filter(Facility.nameplate_capacity != None)
                            .one_or_none()
                        )
                    except MultipleResultsFound:
                        logger.warn(
                            "Multiple results found for duid : {}".format(duid)
                        )

                    if facility:
                        if facility.station and not facility_station:
                            facility_station = facility.station

                        logger.info(
                            "GI: Found facility by DUID: code {} station {}".format(
                                facility.code,
                                facility.station.name
                                if facility.station
                                else None,
                            )
                        )

                # Done trying to find existing
                if not facility:
                    facility = Facility(
                        code=unit_code,
                        network_code=duid,
                        created_by="pipeline.aemo.general_information",
                    )
                    facility.station = facility_station

                    created_facility = True

                if duid and not facility.network_code:
                    facility.network_code = duid
                    facility.updated_by = "pipeline.aemo.general_information"

                if not facility.network_region:
                    facility.network_region = facility_network_region
                    facility.updated_by = "pipeline.aemo.general_information"

                if not facility.network_name:
                    facility.network_name = facility_network_name
                    facility.updated_by = "pipeline.aemo.general_information"

                if not facility.fueltech_id and facility_fueltech:
                    facility.fueltech_id = facility_fueltech
                    facility.updated_by = "pipeline.aemo.general_information"

                if not facility.capacity_registered or (
                    facility.status and facility.status != "operating"
                ):
                    facility.capacity_registered = reg_cap
                    facility.updated_by = "pipeline.aemo.general_information"

                # @TODO work this out
                # facility.dispatch_type = facility_dispatch_type

                if not facility.unit_id:
                    facility.unit_id = unit.id
                    facility.unit_number = unit.number
                    facility.unit_size = unit_size
                    facility.unit_alias = unit.alias

                if not facility.unit_capacity or (
                    facility.status and facility.status != "operating"
                ):
                    facility.unit_capacity = unit_size
                    facility.updated_by = "pipeline.aemo.general_information"

                if not facility.status_id:
                    facility.status_id = facility_status
                    facility.updated_by = "pipeline.aemo.general_information"

                if not facility.registered and facility_comissioned_dt:
                    facility.registered = facility_comissioned_dt
                    facility.updated_by = "pipeline.aemo.general_information"

                facility.station = facility_station

                # if facility.fueltech_id is None:
                #     logger.error(
                #         "Could not find fueltech for: {}".format(item)
                #     )

                # if not facility.nameplate_capacity:
                #     facility.nameplate_capacity = clean_capacity(
                #         item["UpperCapacity"] or item["NameCapacity"]
                #     )

                # facility.status_id = facility_status

                if facility_station and not facility.station:
                    facility.station = facility_station

                if facility.status_id is None:
                    raise Exception(
                        "GI: Failed to map status ({}) on row: {}".format(
                            facility.status_id, facility_record
                        )
                    )

                s.add(facility)
                s.commit()

                facility_index += units_num

                if created_station:
                    logger.info(
                        "GI: {} station with name {} ".format(
                            "Created" if created_station else "Updated",
                            station_name,
                            # facility_station.id,
                        )
                    )

                if created_facility:
                    logger.info(
                        "GI: {} facility with duid {} to station {}".format(
                            "Created" if created_facility else "Updated",
                            duid,
                            station_name,
                        )
                    )

        try:
            s.commit()
        except Exception as e:
            logger.error(e)
            raise e
        finally:
            s.close()

        pass

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        if not "generators" in item and type(item["generators"]) is not list:
            raise Exception("Invalid item - no generators located")

        generators = item["generators"]

        self.process_facilities(generators)

        # with open("test.json", "w") as fh:
        # json.dump(generators, fh, cls=OpenNEMJSONEncoder, indent=4)



"""
    OpenNEM primary schema adapted to support multiple energy sources

    Currently supported:

    - NEM
    - WEM
"""

from decimal import Decimal
from typing import Optional

from geoalchemy2 import Geometry
from shapely import wkb
from sqlalchemy import (
    Boolean,
    Column,
    Date,
    DateTime,
    Enum,
    ForeignKey,
    Index,
    Integer,
    Numeric,
    Sequence,
    String,
    Table,
    Text,
    Time,
    func,
)
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_method, hybrid_property
from sqlalchemy.orm import backref, relationship
from sqlalchemy.sql import func

from opennem.core.dispatch_type import DispatchType
from opennem.core.oid import get_ocode, get_oid

Base = declarative_base()
metadata = Base.metadata


class BaseModel(object):
    """
        Base model for both NEM and WEM

        Each table has an overrid for update and additional meta fields

        @TODO - upsert support for postgresql and mysql dialects (see db/__init__)
    """

    def update(self, **kwargs):
        for key, value in kwargs.items():
            if key == "id" or key.endswith("_id"):
                continue

            # @TODO watch how we do updates so we don't overwrite data
            cur_val = getattr(self, key)
            if key not in [None, ""]:
                setattr(self, key, value)

    def __compile_query(self, query):
        """Via http://nicolascadou.com/blog/2014/01/printing-actual-sqlalchemy-queries"""
        compiler = (
            query.compile
            if not hasattr(query, "statement")
            else query.statement.compile
        )
        return compiler(dialect=postgresql.dialect())

    def __upsert(
        self,
        session,
        model,
        rows,
        as_of_date_col="report_date",
        no_update_cols=[],
    ):
        table = model.__table__

        stmt = self.insert(table).values(rows)

        update_cols = [
            c.name
            for c in table.c
            if c not in list(table.primary_key.columns)
            and c.name not in no_update_cols
        ]

        on_conflict_stmt = stmt.on_conflict_do_update(
            index_elements=table.primary_key.columns,
            set_={k: getattr(stmt.excluded, k) for k in update_cols},
            index_where=(model.report_date < stmt.excluded.report_date),
        )

        print(self.compile_query(on_conflict_stmt))
        session.execute(on_conflict_stmt)

    created_by = Column(Text, nullable=True)
    updated_by = Column(Text, nullable=True)
    processed_at = Column(DateTime(timezone=True), nullable=True)
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())


class FuelTech(Base, BaseModel):
    __tablename__ = "fueltech"

    code = Column(Text, primary_key=True)
    label = Column(Text, nullable=True)
    renewable = Column(Boolean, default=False)

    facilities = relationship("Facility")


class Network(Base, BaseModel):
    __tablename__ = "network"

    code = Column(Text, primary_key=True)
    country = Column(Text, nullable=False)
    label = Column(Text, nullable=True)


class FacilityStatus(Base, BaseModel):
    __tablename__ = "facility_status"

    code = Column(Text, primary_key=True)
    label = Column(Text)


class Participant(Base, BaseModel):
    __tablename__ = "participant"

    id = Column(
        Integer,
        Sequence("seq_participant_id", start=1000, increment=1),
        primary_key=True,
    )

    code = Column(Text, unique=True, index=True)
    name = Column(Text)
    network_name = Column(Text)
    network_code = Column(Text)
    country = Column(Text)
    abn = Column(Text)


class Station(Base, BaseModel):
    __tablename__ = "station"

    def __str__(self):
        return "{} <{}>".format(self.name, self.code)

    def __repr__(self):
        return "{} {} <{}>".format(self.__class__, self.name, self.code)

    id = Column(
        Integer,
        Sequence("seq_station_id", start=1000, increment=1),
        primary_key=True,
    )

    network_id = Column(
        Text,
        ForeignKey("network.code", name="fk_station_network_code"),
        nullable=False,
    )
    network = relationship("Network")

    participant_id = Column(
        Integer,
        ForeignKey("participant.id", name="fk_station_participant_id"),
        nullable=True,
    )
    participant = relationship("Participant")

    facilities = relationship("Facility")

    code = Column(Text, unique=True, index=True, nullable=True)
    name = Column(Text)

    address1 = Column(Text)
    address2 = Column(Text)
    locality = Column(Text)
    state = Column(Text)
    postcode = Column(Text, nullable=True)

    # Original network fields
    network_code = Column(Text, index=True)
    network_name = Column(Text)

    # Geo fields
    place_id = Column(Text, nullable=True, index=True)
    geocode_approved = Column(Boolean, default=False)
    geocode_skip = Column(Boolean, default=False)
    geocode_processed_at = Column(DateTime, nullable=True)
    geocode_by = Column(Text, nullable=True)
    geom = Column(Geometry("POINT", srid=4326))
    boundary = Column(Geometry("MULTIPOLYGON", srid=4326))

    @hybrid_property
    def capacity_registered(self) -> Optional[int]:
        """
            This is the sum of registered capacities for all units for
            this station

        """
        cap_reg = None

        for fac in self.facilities:
            if (
                fac.capacity_registered
                and type(fac.capacity_registered) in [int, float, Decimal,]
                and fac.status_id
                in ["operating", "committed", "commissioning"]
                and fac.dispatch_type == DispatchType.GENERATOR
                and fac.active
            ):
                if not cap_reg:
                    cap_reg = 0

                cap_reg += fac.capacity_registered

        if cap_reg:
            cap_reg = round(cap_reg, 2)

        return cap_reg

    @hybrid_property
    def capacity_aggregate(self) -> Optional[int]:
        """
            This is the sum of aggregate capacities for all units

        """
        cap_agg = None

        for fac in self.facilities:
            if (
                fac.capacity_aggregate
                and type(fac.capacity_aggregate) in [int, float, Decimal,]
                and fac.status_id
                in ["operating", "committed", "commissioning"]
                and fac.dispatch_type == DispatchType.GENERATOR
                and fac.active
            ):
                if not cap_agg:
                    cap_agg = 0

                cap_agg += fac.capacity_aggregate

        if cap_agg:
            cap_agg = round(cap_agg, 2)

        return cap_agg

    @hybrid_property
    def oid(self) -> str:
        return get_oid(self)

    @hybrid_property
    def ocode(self) -> str:
        return get_ocode(self)

    @hybrid_property
    def lat(self) -> Optional[float]:
        if self.geom:
            return wkb.loads(bytes(self.geom.data)).y

        return None

    @hybrid_property
    def lng(self) -> Optional[float]:
        if self.geom:
            return wkb.loads(bytes(self.geom.data)).x

        return None

    @hybrid_property
    def location(self):
        """
            Get lat lng as x,y

            @TODO test this
        """
        if self.geom:
            return (func.as_x(self.geom), func.as_y(self.geom))


class Facility(Base, BaseModel):
    __tablename__ = "facility"

    def __str__(self):
        return "{} <{}>".format(self.code, self.fueltech_id)

    def __repr__(self):
        return "{} {} <{}>".format(self.__class__, self.code, self.fueltech_id)

    id = Column(
        Integer,
        Sequence("seq_facility_id", start=1000, increment=1),
        primary_key=True,
    )

    participant_id = Column(
        Integer,
        ForeignKey("participant.id", name="fk_facility_participant_id"),
        nullable=True,
    )
    participant = relationship("Participant")

    fueltech_id = Column(
        Text,
        ForeignKey("fueltech.code", name="fk_facility_fueltech_id"),
        nullable=True,
    )
    fueltech = relationship("FuelTech", back_populates="facilities")

    status_id = Column(
        Text,
        ForeignKey("facility_status.code", name="fk_facility_status_code"),
    )
    status = relationship("FacilityStatus")

    station_id = Column(
        Integer,
        ForeignKey("station.id", name="fk_station_status_code"),
        nullable=True,
    )
    station = relationship("Station", back_populates="facilities")

    # DUID but modified by opennem as an identifier
    code = Column(Text, nullable=True, index=True, unique=True)

    # Network details
    network_code = Column(Text, nullable=True, index=True)
    network_region = Column(Text, index=True)
    network_name = Column(Text)

    active = Column(Boolean, default=True)

    dispatch_type = Column(
        Enum(DispatchType), nullable=False, default=DispatchType.GENERATOR
    )

    # @TODO remove when ref count is 0
    capacity_registered = Column(Numeric, nullable=True)

    registered = Column(DateTime, nullable=True)
    deregistered = Column(DateTime, nullable=True)

    unit_id = Column(Integer, nullable=True)
    unit_number = Column(Integer, nullable=True)
    unit_alias = Column(Text, nullable=True)
    unit_capacity = Column(Numeric, nullable=True)
    # unit_number_max = Column(Numeric, nullable=True)

    @hybrid_property
    def capacity_aggregate(self) -> Optional[int]:
        """
            This is unit_no * unit_capacity and can differ from registered

        """
        num_units = 1
        cap_aggr = None

        if not self.active:
            return 0

        if self.unit_number and type(self.unit_number) is int:
            num_units = self.unit_number

        if self.unit_capacity and type(self.unit_capacity) is Decimal:
            cap_aggr = num_units * self.unit_capacity

        if type(cap_aggr) is Decimal:
            cap_aggr = round(cap_aggr, 2)

        return cap_aggr

    @hybrid_property
    def duid(self) -> str:
        return self.network_code or self.code

    @hybrid_property
    def status_label(self) -> Optional[str]:
        return self.status.label if self.status else None

    @hybrid_property
    def fueltech_label(self) -> Optional[str]:
        return self.fueltech.label if self.fueltech else None

    @hybrid_property
    def oid(self) -> str:
        return get_oid(self)

    @hybrid_property
    def ocode(self) -> str:
        return get_ocode(self)


class FacilityScada(Base, BaseModel):
    """
        Facility Scada
    """

    __tablename__ = "facility_scada"

    trading_interval = Column(DateTime, index=True, primary_key=True)

    facility_id = Column(
        Integer,
        ForeignKey("facility.id", name="fk_facility_scada_facility_id"),
        primary_key=True,
    )
    facility = relationship("Facility")

    generated = Column(Numeric, nullable=True)
    eoi_quantity = Column(Numeric, nullable=True)


class BalancingSummary(Base, BaseModel):

    __tablename__ = "balancing_summary"

    network_id = Column(
        Text,
        ForeignKey("network.code", name="fk_balancing_summary_network_code"),
    )
    network = relationship("Network")

    trading_interval = Column(DateTime, index=True, primary_key=True)
    forecast_load = Column(Numeric, nullable=True)
    generation_scheduled = Column(Numeric, nullable=True)
    generation_non_scheduled = Column(Numeric, nullable=True)
    generation_total = Column(Numeric, nullable=True)
    price = Column(Numeric, nullable=True)



from typing import Optional

from pydantic import BaseModel, ValidationError, validator

from opennem.core.normalizers import (
    clean_capacity,
    is_number,
    normalize_duid,
    normalize_string,
    participant_name_filter,
    station_name_cleaner,
)

from .opennem import OpennemBaseModel


class OpennemFacility(OpennemBaseModel):
    code: Optional[str]
    name: str
    network_name: str

    @classmethod
    @validator("code")
    def code_clean(cls, v):
        return normalize_duid(v)

    @classmethod
    @validator("network_name")
    def network_name_isalpha(cls, v):
        assert type(v) is str, "must be string"
        return v



import datetime
import re

import pandas as pd
import requests
from sqlalchemy import exc

from dms_daemon import CONFIG, nemweb_requests, orm


def slack_dms_webhook(message):
    json_text = "{0} \n ({1})".format(
        message, datetime.datetime.now().strftime("%Y-%m-%d %H:%M")
    )
    requests.post(
        "https://hooks.slack.com/services/{0}".format(
            CONFIG["slack_hooks"]["error_message"]
        ),
        json={"text": "{0}".format(json_text)},
    )


def table_date(date):
    table = datetime.date(date.year, date.month, 1).strftime("fcas_%Y%m")
    try:
        d = nemweb_requests.last_date(table=table, db="nemweb_causer_pays")
        return d
    except exc.ProgrammingError as E:
        auto_table(E.args[1])
        table_date(date)


def latest_table_date():
    date = datetime.date.today()
    d = table_date(date)

    if d == None:
        date -= datetime.timedelta(20)
        d = table_date(date)

    return d


def process_fcas(link):
    zipIO = nemweb_requests.get_zip(href=link)
    fcas_files = nemweb_requests.ZipFileStreamer(zipIO)
    for fcas_file in fcas_files.filelist:
        yield fcas_files.extract_stream(fcas_file)


def insert_fcas(csvs, table="fcas_201802"):
    connection = orm.engine3.connect()
    transaction = connection.begin()

    for csv in csvs:
        df = pd.read_csv(
            csv,
            parse_dates=[0],
            index_col=[0],
            header=None,
            names=["SETTLEMENTDATE", "ELEMENT_ID", "VARIABLE_ID", "VALUE", "QUALITY"],
        )

        if df.index.dtype.name == "object":
            df.dropna(inplace=True)
            slack_dms_webhook(
                "Issue with FCAS csv (wrong datetime)\n  (CSV file starting at {})".format(
                    df.index[0]
                )
            )

        try:
            df.to_sql(table, con=transaction.connection, if_exists="append")

        except exc.IntegrityError as E:
            transaction.rollback()
            connection.close()
            new_element_keys = check_missing_element(df)
            if new_element_keys != []:
                add_element_keys(new_element_keys)
                insert_fcas(csvs, table=table)
            else:
                raise E

    transaction.commit()
    connection.close()


def check_missing_element(df):
    # could make this into an excpetion?
    unique_elements = list(df.ELEMENT_ID.unique())
    cursor = orm.engine3.execute("SELECT id FROM elements")
    element_keys = [i[0] for i in cursor]
    new_keys = []
    for key in unique_elements:
        if key not in element_keys:
            new_keys.append(key)
    return new_keys


def causer_pays():
    d = latest_table_date()

    page = nemweb_requests.get_page(
        link="http://www.nemweb.com.au/Reports/Current/Causer_Pays/"
    )

    pattern = re.compile("/Reports/Current/Causer_Pays/FCAS_([0-9]{12}).zip")

    for match in pattern.finditer(page.text):
        dt = datetime.datetime.strptime(match.group(1), "%Y%m%d%H%M")
        if dt > d:
            table = dt.strftime("fcas_%Y%m")
            print(match.group(0), table)
            csvs = [csv for csv in process_fcas(match.group(0))]
            insert_fcas(csvs, table=table)


def create_table(table_str="fcas_201801"):
    with orm.engine3.connect() as conn:
        conn.execute("CREATE TABLE {0} like fcas_201802".format(table_str))
        conn.execute(
            "ALTER TABLE `nemweb_causer_pays`.`{0}`  "
            "ADD CONSTRAINT `fk_{0}_1` "
            "FOREIGN KEY (`ELEMENT_ID`) "
            "REFERENCES `nemweb_causer_pays`.`elements` (`ID`) "
            "ON DELETE RESTRICT "
            "ON UPDATE RESTRICT, "
            "ADD CONSTRAINT `fk_{0}_2` "
            "FOREIGN KEY (`VARIABLE_ID`) "
            "REFERENCES `nemweb_causer_pays`.`variables` (`ID`) "
            "ON DELETE RESTRICT "
            "ON UPDATE RESTRICT".format(table_str)
        )


def auto_table(error_msg):
    pattern = re.compile("Table 'nemweb_causer_pays.(fcas_[0-9]{6})' doesn't exist")
    match = pattern.match(error_msg)
    if len(match.groups()) == 1:
        create_table(match.group(1))
        slack_dms_webhook("Created new table: {0}".format(match.group(1)))
    else:
        raise Exception(
            "Auto_table error: wrong table name format", match.group(0).split("'")[1]
        )


def find_new_id(df):
    for element in df.ELEMENT_ID.unique():
        cursor = orm.engine3.execute(
            "SELECT * FROM elements where id = {0}".format(element)
        )
        print(cursor.fetchall()[0])


def read_range():
    d = datetime.datetime(2017, 2, 10, 12, 25)
    for i in range(6):
        d += datetime.timedelta(0, 1800)
        f = archive_file_reader(d)
        csvs = [z for z in process_archive_file(f)]
        insert_fcas(csvs, table="fcas_201711")
        print(d)


def archive_file_reader(d=datetime.datetime(2017, 2, 10, 15, 25)):
    filename = "/data/marble/nemweb/DATA_ARCHIVE/FCAS_Causer_Pays/{0.year}/FCAS_Causer_Pays_{1}/FCAS_{2}.zip".format(
        d, d.strftime("%Y_%m"), d.strftime("%Y%m%d%H%M")
    )
    return filename


def process_archive_file(filename):
    fcas_files = nemweb_requests.ZipFileStreamer(filename)
    for fcas_file in fcas_files.filelist:
        yield fcas_files.extract_stream(fcas_file)


class NewKeyError(KeyError):
    def __init__(self, *args, **kwargs):
        KeyError.__init__(self, *args, **kwargs)

    def __str__(self):
        return "New element id: {0}".format(*self.args)


def add_element_keys(new_element_keys):
    """Adds list of new element keys to elements table"""
    df = pd.DataFrame(new_element_keys, columns=["ID"])
    with orm.engine3.connect() as conn:
        df.to_sql("elements", con=conn, if_exists="append", index=False)
    slack_dms_webhook("Added new elements: {0}".format(new_element_keys))



import datetime
import glob
from time import time
from zipfile import BadZipFile

from sqlalchemy import create_engine, update
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import Session, sessionmaker

from dms_daemon import CONFIG, nemweb_reader
from dms_daemon.settings import get_mysql_host

# Initialises connections to databases
engine = create_engine(get_mysql_host("nemweb"))
engine2 = create_engine(get_mysql_host("nemweb_live"))
engine3 = create_engine(get_mysql_host("nemweb_causer_pays"))
engine_meta = create_engine(get_mysql_host("nemweb_meta"))
engine_derived = create_engine(get_mysql_host("nemweb_derived"))

print("meta host", get_mysql_host(db_name="nemweb_meta"))

Session = sessionmaker(bind=engine)

Base = automap_base()
Base.prepare(engine, reflect=False)


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


def load_keys(key):
    q = engine.execute("SELECT {0},ID FROM {0}".format(key))
    return AttrDict(q.fetchall())


keys = AttrDict(
    {i: load_keys(i) for i in ["REGIONID", "INTERCONNECTORID", "BIDTYPE", "RUNTYPE"]}
)
keys["DUID"] = AttrDict(
    engine_meta.execute("SELECT {0},ID FROM {0}".format("DUID")).fetchall()
)

region_dict = keys["REGIONID"]
duid_dict = keys["DUID"]
interconnector_dict = keys["INTERCONNECTORID"]

q = engine_meta.execute("SELECT ID,STATION_NAME FROM STATION_NAMES")
station_dict = dict(q.fetchall())

########################################################################################################


class Archive:
    def __init__(self, archive_name):
        self.name = archive_name
        self.filelist = sorted(
            glob.glob("/data/marble/nemweb/ARCHIVE/{}/*".format(archive_name))
        )
        self.__load_table_data()
        self.skip_files = []

    def unzip_archive(self, archive_file):
        for zf_name, zipfile in nemweb_reader.zip_streams(archive_file):
            yield zf_name, zipfile

    def unzip_zipfile(self, fileobject):
        return nemweb_reader.load_nemfile(f=fileobject)

    def __load_table_data(self):
        _, sample_file = next(self.unzip_archive(self.filelist[0]))
        nemfile = self.unzip_zipfile(sample_file)
        self.tables = {table: nemfile.__dict__[table].columns for table in nemfile}


########################################################################################################

"""Posibly have an archive file object??"""

########################################################################################################


class ArchiveInserter(Archive):
    """Archive subclass, with methods to insert tables from the archive"""

    def __init__(self, archive_name="Next_Day_Dispatch"):
        super.__init__(self, archive_name)
        self.csv_table = "DISPATCH_UNIT_SOLUTION"
        self.cols = Base.classes["DISPATCH_UNIT_SOLUTION"].__table__.columns.keys()
        self.cols.remove("ID")

    def insert_table(self):
        for archive_file in self.filelist:
            print(archive_file)
            self.insert_archive_file(archive_file)

    def insert_archive_file(self, archive_file):
        for zf_name, zipfile in self.unzip_archive(archive_file):
            nemfile = self.unzip_zipfile(zipfile)

            df = nemfile.__dict__[self.csv_table].copy()
            self.insert_dataframe(df)

    def insert_dataframe(self, df, map_from="DUID", map_dict=duid_dict):
        # maps duid names to duid ids
        df[map_from] = df[map_from].apply(lambda x: map_dict[x])

        df[self.cols].to_sql(
            name="DISPATCH_UNIT_SOLUTION", index=None, con=engine, if_exists="append"
        )


########################################################################################################


class ArchiveUpdater(Archive):
    """Archive subclass, with methods update tables from the archive"""

    csv_table = None

    def __init__(self, archive_name="DispatchIS_Reports"):
        super.__init__(self, archive_name)

    def update_table(self, map_from="REGIONID", map_dict=region_dict):
        for archive_file in self.filelist:
            print(archive_file)
            self.update_archive_file(archive_file, map_from=map_from, map_dict=map_dict)

    def update_archive_file(
        self, archive_file, map_from="REGIONID", map_dict=region_dict
    ):
        session = Session()

        for zf_name, zipfile in self.unzip_archive(archive_file):

            # block for dealing with htm files
            try:
                nemfile = self.unzip_zipfile(zipfile)
            except nemweb_reader.BadZipfile:
                if zf_name.split(".")[-1] == "htm":
                    continue
                elif zf_name in self.skip_files:
                    continue
                else:
                    continue

            df = nemfile.__dict__[self.csv_table].copy()

            # maps region names to regionids
            df[map_from] = df[map_from].apply(lambda x: map_dict[x])

            for i, row in df.iterrows():
                sql = self.row_update(row)
                session.execute(sql)
            session.commit()

        session.close()

    def row_update(self, row):

        sql = update(self.db_table)
        for i in self.index_columns:
            sql = sql.where(self.db_table.__dict__[i] == row[i])
        return sql.values(row[self.update_columns].to_dict())

    def legacy_filelist(self):
        self.filelist = []
        for year in range(2006, 2012):
            self.filelist += glob.glob(
                "/data/marble/nemweb/LEGACY/public_dispatchis/public_dispatchis_{0}/*.zip".format(
                    year
                )
            )


########################################################################################################


class LegacyArchiveUpdater(ArchiveUpdater):
    def __init__(self, archive_name="DispatchIS_Reports"):
        Archive.__init__(self, archive_name)

    # adds anther unzip step to update function
    def update_table(self):
        for zipfile in self.filelist:
            for zf_name, archive_file in self.unzip_archive(zipfile):
                print(zf_name)
                self.update_archive_file(archive_file)

    def legacy_filelist(self, year=2005):
        self.filelist = glob.glob(
            "/data/marble/nemweb/LEGACY/public_dispatchis/public_dispatchis_{0}/*.zip".format(
                year
            )
        )


########################################################################################################
# Instances of updaters


class trading_price_updater(ArchiveUpdater):
    def __init__(self, archive_name="TradingIS_Reports"):
        ArchiveUpdater.__init__(self, archive_name)
        self.csv_table = "PRICE"
        self.db_table = Base.classes["TRADING_PRICE"]
        self.index_columns = ["SETTLEMENTDATE", "REGIONID"]
        self.update_columns = [
            "ROP",
            "RAISE6SECRRP",
            "RAISE6SECROP",
            "RAISE60SECRRP",
            "RAISE60SECROP",
            "RAISE5MINRRP",
            "RAISE5MINROP",
            "RAISEREGRRP",
            "RAISEREGROP",
            "LOWER6SECRRP",
            "LOWER6SECROP",
            "LOWER60SECRRP",
            "LOWER60SECROP",
            "LOWER5MINRRP",
            "LOWER5MINROP",
            "LOWERREGRRP",
            "LOWERREGROP",
        ]


class trading_regionsum_updater(ArchiveUpdater):
    def __init__(self, archive_name="TradingIS_Reports"):
        ArchiveUpdater.__init__(self, archive_name)
        self.csv_table = "REGIONSUM"
        self.db_table = Base.classes["TRADING_REGIONSUM"]
        self.index_columns = ["SETTLEMENTDATE", "REGIONID"]
        self.update_columns = [
            "LOWERREGLOCALDISPATCH",
            "LOWER5MINLOCALDISPATCH",
            "LOWER60SECLOCALDISPATCH",
            "LOWER6SECLOCALDISPATCH",
            "RAISE5MINLOCALDISPATCH",
            "RAISEREGLOCALDISPATCH",
            "RAISE60SECLOCALDISPATCH",
            "RAISE6SECLOCALDISPATCH",
        ]


class dispatch_price_updater(ArchiveUpdater):
    def __init__(self, archive_name="DispatchIS_Reports"):
        ArchiveUpdater.__init__(self, archive_name)
        self.csv_table = "PRICE"
        self.db_table = Base.classes["DISPATCH_PRICE"]
        self.index_columns = ["SETTLEMENTDATE", "REGIONID", "INTERVENTION"]
        self.update_columns = [
            "ROP",
            "APCFLAG",
            "RAISE6SECRRP",
            "RAISE6SECROP",
            "RAISE6SECAPCFLAG",
            "RAISE60SECRRP",
            "RAISE60SECROP",
            "RAISE60SECAPCFLAG",
            "RAISE5MINRRP",
            "RAISE5MINROP",
            "RAISE5MINAPCFLAG",
            "RAISEREGRRP",
            "RAISEREGROP",
            "RAISEREGAPCFLAG",
            "LOWER6SECRRP",
            "LOWER6SECROP",
            "LOWER6SECAPCFLAG",
            "LOWER60SECRRP",
            "LOWER60SECROP",
            "LOWER60SECAPCFLAG",
            "LOWER5MINRRP",
            "LOWER5MINROP",
            "LOWER5MINAPCFLAG",
            "LOWERREGRRP",
            "LOWERREGROP",
            "LOWERREGAPCFLAG",
        ]


class dispatch_regionsum_updater(ArchiveUpdater):
    def __init__(self, archive_name="DispatchIS_Reports"):
        ArchiveUpdater.__init__(self, archive_name)
        self.csv_table = "REGIONSUM"
        self.db_table = Base.classes["DISPATCH_REGIONSUM"]
        self.index_columns = ["SETTLEMENTDATE", "REGIONID", "INTERVENTION"]
        self.update_columns = [
            "LOWERREGLOCALDISPATCH",
            "LOWER5MINLOCALDISPATCH",
            "LOWER60SECLOCALDISPATCH",
            "LOWER6SECLOCALDISPATCH",
            "RAISE5MINLOCALDISPATCH",
            "RAISEREGLOCALDISPATCH",
            "RAISE60SECLOCALDISPATCH",
            "RAISE6SECLOCALDISPATCH",
            "LOWERREGACTUALAVAILABILITY",
            "LOWER5MINACTUALAVAILABILITY",
            "LOWER60SECACTUALAVAILABILITY",
            "LOWER6SECACTUALAVAILABILITY",
            "RAISE5MINACTUALAVAILABILITY",
            "RAISEREGACTUALAVAILABILITY",
            "RAISE60SECACTUALAVAILABILITY",
            "RAISE6SECACTUALAVAILABILITY",
            "AGGREGATEDISPATCHERROR",
        ]


class predispatch_price_updater(ArchiveUpdater):
    def __init__(self, archive_name="PreDispatchIS_Reports"):
        ArchiveUpdater.__init__(self, archive_name)
        self.csv_table = "REGION_PRICES"
        self.db_table = Base.classes["PREDISPATCH_REGION_PRICES"]
        self.index_columns = ["PREDISPATCHSEQNO", "INTERVENTION", "REGIONID", "PERIODID"]
        self.update_columns = [
            "RAISE6SECRRP",
            "RAISE60SECRRP",
            "RAISE5MINRRP",
            "RAISEREGRRP",
            "LOWER6SECRRP",
            "LOWER60SECRRP",
            "LOWER5MINRRP",
            "LOWERREGRRP",
        ]
        self.skip_files += [
            "PUBLIC_PREDISPATCHIS_201610121900_20161012183215.zip",
            "PUBLIC_PREDISPATCHIS_201610130000_20161012233150.zip",
        ]


class bidday_updater(ArchiveUpdater):
    def __init__(self, archive_name="Next_Day_Offer_Energy"):
        ArchiveUpdater.__init__(self, archive_name)
        self.csv_table = "BIDDAYOFFER"
        self.db_table = Base.classes["OFFER_BIDDAYOFFER"]
        self.index_columns = [
            "SETTLEMENTDATE",
            "DUID",
            "OFFERDATE",
            "VERSIONNO",
            "ENTRYTYPE",
            "LASTCHANGED",
        ]
        self.update_columns = ["T1", "T2", "T3", "T4"]


# ==============================Dealing with LIVE data (one less level in archive)
class ScadaInserter(Archive):
    def __init__(self, live_name):
        self.name = live_name
        self.filelist = sorted(
            glob.glob("/data/marble/nemweb/LIVE/{}/*".format(live_name))
        )

    def load_table_data(self):
        for f in self.filelist:
            self.nemfile = nemweb_reader.load_nemfile(f)
            self.key_map()

    def key_map(self, map_from="DUID", map_dict=duid_dict):
        for table in self.nemfile:
            df = self.nemfile.__dict__[table]
            df[map_from] = df[map_from].apply(lambda x: map_dict[x])
            df = df[["SETTLEMENTDATE", "DUID", "SCADAVALUE"]]
            df.to_sql("DISPATCH_UNIT_SCADA", con=engine2, if_exists="append", index=None)


class RegionPriceInserter(Archive):
    def __init__(self, live_name="DispatchIS_Reports"):
        self.name = live_name
        self.filelist = sorted(
            glob.glob("/data/marble/nemweb/LIVE/{}/*".format(live_name))
        )

    def load_table_data(self):
        for f in self.filelist:
            self.nemfile = nemweb_reader.load_nemfile(f)
            self.key_map()

    def key_map(self, map_from="REGIONID", map_dict=region_dict):
        for table in ["DISPATCH_PRICE"]:
            df = self.nemfile.__dict__[table]
            df[map_from] = df[map_from].apply(lambda x: map_dict[x])
            df = df[
                [
                    "SETTLEMENTDATE",
                    "REGIONID",
                    "DISPATCHINTERVAL",
                    "RRP",
                    "ROP",
                    "APCFLAG",
                    "CUMUL_PRE_AP_ENERGY_PRICE",
                    "INTERVENTION",
                    "RAISE6SECRRP",
                    "RAISE6SECROP",
                    "RAISE6SECAPCFLAG",
                    "RAISE60SECRRP",
                    "RAISE60SECROP",
                    "RAISE60SECAPCFLAG",
                    "RAISE5MINRRP",
                    "RAISE5MINROP",
                    "RAISE5MINAPCFLAG",
                    "RAISEREGRRP",
                    "RAISEREGROP",
                    "RAISEREGAPCFLAG",
                    "LOWER6SECRRP",
                    "LOWER6SECROP",
                    "LOWER6SECAPCFLAG",
                    "LOWER60SECRRP",
                    "LOWER60SECROP",
                    "LOWER60SECAPCFLAG",
                    "LOWER5MINRRP",
                    "LOWER5MINROP",
                    "LOWER5MINAPCFLAG",
                    "LOWERREGRRP",
                    "LOWERREGROP",
                    "LOWERREGAPCFLAG",
                ]
            ]
            df.to_sql("DISPATCH_PRICE", con=engine2, if_exists="append", index=None)


class RegionSum(Archive):
    def __init__(self, live_name="DispatchIS_Reports"):
        self.name = live_name
        self.filelist = sorted(
            glob.glob("/data/marble/nemweb/LIVE/{}/*".format(live_name))
        )

    def load_table_data(self):
        for f in self.filelist:
            self.nemfile = nemweb_reader.load_nemfile(f)
            self.key_map()

    def key_map(self, map_from="REGIONID", map_dict=region_dict):
        for table in ["DISPATCH_REGIONSUM"]:
            df = self.nemfile.__dict__[table]
            df[map_from] = df[map_from].apply(lambda x: map_dict[x])
            df = df[
                [
                    "SETTLEMENTDATE",
                    "REGIONID",
                    "TOTALDEMAND",
                    "AVAILABLEGENERATION",
                    "AVAILABLELOAD",
                    "DEMANDFORECAST",
                    "DISPATCHABLEGENERATION",
                    "DISPATCHABLELOAD",
                    "NETINTERCHANGE",
                    "EXCESSGENERATION",
                    "INITIALSUPPLY",
                    "CLEAREDSUPPLY",
                    "TOTALINTERMITTENTGENERATION",
                    "DEMAND_AND_NONSCHEDGEN",
                    "UIGF",
                    "SEMISCHEDULE_CLEAREDMW",
                    "DISPATCHINTERVAL",
                    "SEMISCHEDULE_COMPLIANCEMW",
                    "INTERVENTION",
                    "LOWERREGLOCALDISPATCH",
                    "LOWER5MINLOCALDISPATCH",
                    "LOWER60SECLOCALDISPATCH",
                    "LOWER6SECLOCALDISPATCH",
                    "RAISE5MINLOCALDISPATCH",
                    "RAISEREGLOCALDISPATCH",
                    "RAISE60SECLOCALDISPATCH",
                    "RAISE6SECLOCALDISPATCH",
                    "AGGREGATEDISPATCHERROR",
                    "RAISE6SECACTUALAVAILABILITY",
                    "RAISE5MINACTUALAVAILABILITY",
                    "RAISE60SECACTUALAVAILABILITY",
                    "RAISEREGACTUALAVAILABILITY",
                    "LOWER6SECACTUALAVAILABILITY",
                    "LOWER5MINACTUALAVAILABILITY",
                    "LOWER60SECACTUALAVAILABILITY",
                    "LOWERREGACTUALAVAILABILITY",
                ]
            ]
            df.to_sql("DISPATCH_REGIONSUM", con=engine2, if_exists="append", index=None)


class TradingRegionSum(Archive):
    def __init__(self, live_name="TradingIS_Reports"):
        self.name = live_name
        self.filelist = sorted(
            glob.glob("/data/marble/nemweb/LIVE/{}/*".format(live_name))
        )

    def load_table_data(self):
        for f in self.filelist:
            self.nemfile = nemweb_reader.load_nemfile(f)
            self.key_map()

    def key_map(self, map_from="REGIONID", map_dict=region_dict):
        for table in ["TRADING_REGIONSUM"]:
            df = self.nemfile.__dict__[table]
            df[map_from] = df[map_from].apply(lambda x: map_dict[x])
            df = df[
                [
                    "SETTLEMENTDATE",
                    "REGIONID",
                    "TOTALDEMAND",
                    "AVAILABLEGENERATION",
                    "AVAILABLELOAD",
                    "DEMANDFORECAST",
                    "DISPATCHABLEGENERATION",
                    "DISPATCHABLELOAD",
                    "NETINTERCHANGE",
                    "EXCESSGENERATION",
                    "INITIALSUPPLY",
                    "CLEAREDSUPPLY",
                    "TOTALINTERMITTENTGENERATION",
                    "DEMAND_AND_NONSCHEDGEN",
                    "UIGF",
                    "LOWERREGLOCALDISPATCH",
                    "LOWER5MINLOCALDISPATCH",
                    "LOWER60SECLOCALDISPATCH",
                    "LOWER6SECLOCALDISPATCH",
                    "RAISE5MINLOCALDISPATCH",
                    "RAISEREGLOCALDISPATCH",
                    "RAISE60SECLOCALDISPATCH",
                    "RAISE6SECLOCALDISPATCH",
                ]
            ]
            df.to_sql("TRADING_REGIONSUM", con=engine2, if_exists="append", index=None)


class TradingPrice(Archive):
    def __init__(self, live_name="TradingIS_Reports"):
        self.name = live_name
        self.filelist = sorted(
            glob.glob("/data/marble/nemweb/LIVE/{}/*".format(live_name))
        )

    def load_table_data(self):
        for f in self.filelist:
            self.nemfile = nemweb_reader.load_nemfile(f)
            self.key_map()

    def key_map(self, map_from="REGIONID", map_dict=region_dict):
        for table in ["TRADING_PRICE"]:
            df = self.nemfile.__dict__[table]
            df[map_from] = df[map_from].apply(lambda x: map_dict[x])
            df = df[
                [
                    "SETTLEMENTDATE",
                    "REGIONID",
                    "RRP",
                    "ROP",
                    "RAISE6SECRRP",
                    "RAISE6SECROP",
                    "RAISE60SECRRP",
                    "RAISE60SECROP",
                    "RAISE5MINRRP",
                    "RAISE5MINROP",
                    "RAISEREGRRP",
                    "RAISEREGROP",
                    "LOWER6SECRRP",
                    "LOWER6SECROP",
                    "LOWER60SECRRP",
                    "LOWER60SECROP",
                    "LOWER5MINRRP",
                    "LOWER5MINROP",
                    "LOWERREGRRP",
                    "LOWERREGROP",
                ]
            ]
            df.to_sql("TRADING_PRICE", con=engine2, if_exists="append", index=None)


###############################################################################3
class DispatchUnitSolution(Archive):
    def __init__(self, name="Next_Day_Dispatch"):
        self.name = name
        self.filelist = sorted(
            glob.glob("/data/marble/nemweb/ARCHIVE/{0}/*".format(name))
        )

    def load_table_data(self):
        for f in self.filelist:
            print(f)
            self.nemfile = nemweb_reader.load_nemfile(f)
            self.key_map()

            break

    def key_map(self, map_from="DUID", map_dict=duid_dict):
        for table in ["DISPATCH_UNIT_SOLUTION"]:
            df = self.nemfile.__dict__[table]
            df[map_from] = df[map_from].apply(lambda x: map_dict[x])
            break
            df = df[
                [
                    "SETTLEMENTDATE",
                    "REGIONID",
                    "RRP",
                    "ROP",
                    "RAISE6SECRRP",
                    "RAISE6SECROP",
                    "RAISE60SECRRP",
                    "RAISE60SECROP",
                    "RAISE5MINRRP",
                    "RAISE5MINROP",
                    "RAISEREGRRP",
                    "RAISEREGROP",
                    "LOWER6SECRRP",
                    "LOWER6SECROP",
                    "LOWER60SECRRP",
                    "LOWER60SECROP",
                    "LOWER5MINRRP",
                    "LOWER5MINROP",
                    "LOWERREGRRP",
                    "LOWERREGROP",
                ]
            ]
            # df.to_sql('TRADING_PRICE',con=engine, if_exists= 'append',index=None)


###############################################################################3
class PreDispatchSensitivities(ArchiveInserter):
    def __init__(self, name="Predispatch_Sensitivities"):
        self.name = name
        self.filelist = sorted(
            glob.glob("/data/marble/nemweb/ARCHIVE/{0}/*".format(name))
        )
        # self.filelist = self.filelist[280:]
        # self.filelist = self.filelist[330:]
        self.filelist = self.filelist[368:]

        self.csv_table = "PREDISPATCH_PRICESENSITIVITIES"
        self.cols = Base.classes[
            "PREDISPATCH_PRICESENSITIVITIES"
        ].__table__.columns.keys()
        self.cols.remove("ID")

    def insert_dataframe(self, df, map_from="REGIONID", map_dict=region_dict):
        # maps duid names to duid ids
        df[map_from] = df[map_from].apply(lambda x: map_dict[x])

        df[self.cols].to_sql(
            name="PREDISPATCH_PRICESENSITIVITIES",
            index=None,
            con=engine,
            if_exists="append",
        )

    def insert_archive_file(self, archive_file):
        for zf_name, zipfile in self.unzip_archive(archive_file):
            try:
                nemfile = self.unzip_zipfile(zipfile)
                df = nemfile.__dict__[self.csv_table].copy()
                self.insert_dataframe(df)
            except BadZipFile as E:
                print(zf_name)
                if zf_name in [
                    "PUBLIC_PREDISPATCH_SENSITIVITIES_20161227093218_0000000279137718.zip",
                    "PUBLIC_PREDISPATCH_SENSITIVITIES_20170913020303_0000000287276933.zip",
                ]:
                    pass
                else:
                    raise E



import json
import logging
from datetime import datetime
from typing import Optional

from scrapy.exceptions import DropItem
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy.sql import text

from opennem.core.dispatch_type import DispatchType, parse_dispatch_type
from opennem.core.facilitystations import facility_station_join_by_name
from opennem.core.facilitystatus import map_v3_states
from opennem.core.fueltechs import lookup_fueltech
from opennem.core.normalizers import (
    clean_capacity,
    name_normalizer,
    normalize_aemo_region,
    normalize_duid,
    normalize_string,
    participant_name_filter,
    station_name_cleaner,
)
from opennem.core.station_duid_map import (
    facility_has_station_remap,
    facility_map_station,
)
from opennem.db.models.opennem import Facility, FacilityStatus
from opennem.db.models.opennem import Participant as ParticipantModel
from opennem.db.models.opennem import Station
from opennem.pipelines import DatabaseStoreBase
from opennem.schema.pariticipant import OpennemParticipant
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class NemMMSSingle(DatabaseStoreBase):
    """

    """

    def get_table(self, item):
        if not "tables" in item:
            logger.error(item)
            raise Exception("No tables passed to pipeline")

        table_names = [i["name"] for i in item["tables"]]

        if not self.table in table_names:
            logger.debug(
                "Skipping {} pipeline step as table {} not processed".format(
                    self.__class__, self.table
                )
            )
            return False

        table = [
            i
            for i in item["tables"]
            if "name" in i and i["name"] == self.table
        ]

        return table.pop() if len(table) else None


class NemStoreMMSStations(DatabaseStoreBase):
    """

    """

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        records_updated = 0
        records_created = 0

        for record in item:
            created = False

            duid = normalize_duid(record["STATIONID"])
            name = station_name_cleaner(record["STATIONNAME"])
            network_name = normalize_string(record["STATIONNAME"])
            address1 = normalize_string(record["ADDRESS1"])
            address2 = normalize_string(record["ADDRESS2"])
            city = normalize_string(record["CITY"])
            state = normalize_string(record["STATE"]).capitalize()
            postcode = normalize_string(record["POSTCODE"])

            station = (
                s.query(Station)
                .filter(Station.network_code == duid)
                .one_or_none()
            )

            if not station:
                station = Station(
                    code=duid,
                    network_code=duid,
                    created_by="au.nem.mms.stations",
                )

                records_created += 1
                created = True
            else:
                station.updated_by = "au.nem.mms.stations"
                records_updated += 1

            station.name = name
            station.network_id = "NEM"
            station.network_name = network_name
            station.address1 = address1
            station.address2 = address2
            station.locality = city
            station.state = state
            station.postcode = postcode

            try:
                s.add(station)
                s.commit()
            except Exception as e:
                logger.error(e)

            logger.debug(
                "{} station record with id {}".format(
                    "Created" if created else "Updated", duid
                )
            )

        logger.info(
            "Created {} records and updated {}".format(
                records_created, records_updated
            )
        )


class NemStoreMMSStationStatus(DatabaseStoreBase):
    """

    """

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        for record in item:
            duid = normalize_duid(record["STATIONID"])
            # authorized_date = name_normalizer(record["AUTHORISEDDATE"])
            status = map_v3_states(record["STATUS"])

            station = (
                s.query(Station)
                .filter(Station.network_code == duid)
                .one_or_none()
            )

            if not station:
                logger.error("Could not find station {}".format(duid))
                continue

            # @TODO station statuses -> facilities should be set to retired if active

            try:
                s.add(station)
                s.commit()
            except Exception as e:
                logger.error(e)


class NemStoreMMSParticipant(DatabaseStoreBase):
    """

        @NOTE This pipeline has been converted to use pydantic models
    """

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        records_updated = 0
        records_created = 0

        q = self.engine.execute(text("select code from participant"))

        participant_codes = list(set([i[0] for i in q.fetchall()]))

        records = item
        for record in records:
            created = False

            participant_schema = OpennemParticipant(
                **{
                    "code": record["PARTICIPANTID"],
                    "name": record["NAME"],
                    "network_name": record["NAME"],
                }
            )

            # pid = normalize_duid(record["PARTICIPANTID"])
            # name = normalize_string(record["NAME"])
            # name_clean = participant_name_filter(record["NAME"])

            participant = (
                s.query(ParticipantModel)
                .filter(ParticipantModel.code == participant_schema.code)
                .one_or_none()
            )

            if not participant:
                participant = ParticipantModel(
                    **{
                        **participant_schema.dict(),
                        "created_by": "au.nem.mms.participant",
                    }
                )

                records_created += 1
                created = True
            else:
                participant.name = participant_schema.name
                participant.network_name = participant_schema.network_name
                records_updated += 1

            try:
                s.add(participant)
                s.commit()
            except Exception as e:
                logger.error(e)

            logger.debug(
                "{} participant record with id {}".format(
                    "Created" if created else "Updated",
                    participant_schema.code,
                )
            )

        logger.info(
            "Created {} records and updated {}".format(
                records_created, records_updated
            )
        )


class NemStoreMMSDudetail(DatabaseStoreBase):
    """

    """

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        records_updated = 0
        records_created = 0

        for record in item:
            created = False

            duid = normalize_duid(record["DUID"])
            capacity_registered = clean_capacity(record["REGISTEREDCAPACITY"])
            capacity_max = clean_capacity(record["MAXCAPACITY"])
            dispatch_type = parse_dispatch_type(record["DISPATCHTYPE"])

            facility = (
                s.query(Facility)
                .filter(Facility.network_code == duid)
                .one_or_none()
            )

            if not facility:
                facility = Facility(
                    code=duid,
                    network_code=duid,
                    status_id="retired",
                    dispatch_type=dispatch_type,
                    created_by="au.nem.mms.dudetail",
                )

                records_created += 1
                created = True
            else:
                facility.updated_by = "au.nem.mms.dudetail"
                records_updated += 1

            facility.capacity_registered = capacity_registered
            facility.capacity_max = capacity_max

            try:
                s.add(facility)
                s.commit()
            except Exception as e:
                logger.error(e)

            logger.debug(
                "{} facility record with id {}".format(
                    "Created" if created else "Updated", duid
                )
            )

        logger.info(
            "Created {} facility records and updated {}".format(
                records_created, records_updated
            )
        )


class NemStoreMMSDudetailSummary(DatabaseStoreBase):
    """

    """

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        records_updated = 0
        records_created = 0

        for record in item:
            created = False

            participant_code = normalize_duid(
                record["facilities"][0]["PARTICIPANTID"]
            )
            dispatch_type = parse_dispatch_type(record["DISPATCHTYPE"])

            # Step 1. Find participant by code or create
            participant = (
                s.query(ParticipantModel)
                .filter(ParticipantModel.code == participant_code)
                .one_or_none()
            )

            if not participant:
                participant = ParticipantModel(
                    code=participant_code,
                    network_code=participant_code,
                    created_by="au.nem.mms.dudetail_summary",
                )
                logger.debug("Created participant {}".format(participant_code))
            else:
                participant.updated_by = "au.nem.mms.dudetail_summary"

            # Step 3. now create the facilities and associate
            for facility_record in record["facilities"]:
                duid = normalize_duid(facility_record["DUID"])
                station_code = facility_map_station(
                    duid, normalize_duid(record["id"])
                )

                network_region = normalize_aemo_region(
                    facility_record["REGIONID"]
                )
                date_start = facility_record["date_start"]
                date_end = facility_record["date_end"]
                facility_state = "retired"

                # Step 2. Find station or create
                station = (
                    s.query(Station)
                    .filter(Station.network_code == station_code)
                    .one_or_none()
                )

                if not station:
                    station = Station(
                        code=station_code,
                        network_code=station_code,
                        network_id="NEM",
                        created_by="au.nem.mms.dudetail_summary",
                    )
                    logger.debug("Created station {}".format(station_code))
                else:
                    station.updated_by = "au.nem.mms.dudetail_summary"

                station.participant = participant

                if date_end == None:
                    facility_state = "operating"

                facility = (
                    s.query(Facility)
                    .filter(Facility.network_code == duid)
                    .one_or_none()
                )

                if not facility:
                    facility = Facility(
                        code=duid,
                        network_code=duid,
                        dispatch_type=dispatch_type,
                        created_by="au.nem.mms.dudetail_summary",
                    )

                    records_created += 1
                    created = True
                else:
                    facility.updated_by = "au.nem.mms.dudetail_summary"
                    records_updated += 1

                facility.network_region = network_region
                facility.deregistered = date_end
                facility.registered = date_start

                facility.status_id = facility_state

                # Associations
                facility_station_id = facility_map_station(duid, station.id)

                facility.station_id = station.id

            try:
                s.add(facility)
                s.commit()
            except Exception as e:
                logger.error(e)

            logger.debug(
                "{} facility record with id {}".format(
                    "Created" if created else "Updated", duid
                )
            )

        logger.info(
            "Created {} facility records and updated {}".format(
                records_created, records_updated
            )
        )


class NemStoreMMSStatdualloc(DatabaseStoreBase):
    """
        AEMO MMS associates all duids with station ids
    """

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        records_updated = 0
        records_created = 0

        for record in item:
            created = False

            duid = normalize_duid(record["DUID"])
            station_code = facility_map_station(
                duid, normalize_duid(record["STATIONID"])
            )

            station = (
                s.query(Station)
                .filter(Station.network_code == station_code)
                .one_or_none()
            )

            facility = (
                s.query(Facility)
                .filter(Facility.network_code == duid)
                .one_or_none()
            )

            if not station:
                station = Station(
                    code=station_code,
                    network_code=station_code,
                    network_id="NEM",
                    created_by="au.nem.mms.statdualloc",
                )

            if not facility:
                facility = Facility(
                    code=duid,
                    network_code=duid,
                    network_id="NEM",
                    status_id="retired",
                    created_by="au.nem.mms.statdualloc",
                )

                records_created += 1
                created = True
            else:
                facility.updated_by = "au.nem.mms.statdualloc"
                records_updated += 1

            facility.station = station

            try:
                s.add(facility)
                s.commit()
            except Exception as e:
                logger.error(e)

            logger.debug(
                "{} facility record with id {}".format(
                    "Created" if created else "Updated", duid
                )
            )

        logger.info(
            "Created {} facility records and updated {}".format(
                records_created, records_updated
            )
        )



import decimal
import json
import os
from datetime import datetime
from pprint import pprint

from smart_open import open

from opennem.api.controllers import (
    bom_observation,
    wem_demand,
    wem_energy_all,
    wem_energy_year,
    wem_market_value_all,
    wem_market_value_year,
    wem_power_groups,
    wem_price,
)

BASE_EXPORT = "s3://data.opennem.org.au"

BASE_EXPORT_LOCAL = os.path.realpath(
    os.path.join(os.path.dirname(__file__), "..", "..", "data")
)


class NemEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)
        if isinstance(o, datetime):
            return o.isoformat()
        return super(NemEncoder, self).default(o)


UPLOAD_ARGS = {
    "ContentType": "application/json",
}


def wem_export_power():
    json_data = wem_power_groups()

    tempratures = bom_observation("009021")
    price = wem_price()
    demand = wem_demand()

    json_data.append(tempratures)
    json_data.append(price)
    json_data.append(demand)

    power_path = BASE_EXPORT + "/power/wem.json"

    with open(
        power_path,
        "w",
        transport_params=dict(multipart_upload_kwargs=UPLOAD_ARGS),
    ) as fh:
        json.dump(json_data, fh, cls=NemEncoder)


def wem_export_years():

    for year in [2020, 2019]:
        json_envelope = []

        energy = wem_energy_year(year)
        market_value = wem_market_value_year(year)

        json_envelope = energy + market_value

        year_path = BASE_EXPORT + f"/wem/energy/daily/{year}.json"

        with open(
            year_path,
            "w",
            transport_params=dict(multipart_upload_kwargs=UPLOAD_ARGS),
        ) as fh:
            json.dump(json_envelope, fh, cls=NemEncoder)


def wem_export_all():

    energy = wem_energy_all()
    market_value = wem_market_value_all()

    json_envelope = energy + market_value

    all_path = BASE_EXPORT + f"/wem/energy/monthly/all.json"

    with open(
        all_path,
        "w",
        transport_params=dict(multipart_upload_kwargs=UPLOAD_ARGS),
    ) as fh:
        json.dump(json_envelope, fh, cls=NemEncoder)


def wem_run_all():
    wem_export_power()
    wem_export_years()
    wem_export_all()


if __name__ == "__main__":
    # j = wem_market_value_year()
    # pprint(json.dumps(j, cls=NemEncoder))

    wem_export_power()
    wem_export_years()
    wem_export_all()



from sqlalchemy import (
    Boolean,
    Column,
    Date,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    Numeric,
    Sequence,
    String,
    Table,
    Text,
    Time,
)
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func

Base = declarative_base()
metadata = Base.metadata


class BomStation(Base):
    __tablename__ = "bom_station"

    code = Column(Text, primary_key=True)
    state = Column(Text)
    name = Column(Text)
    registered = Column(Date)
    lat = Column(Numeric)
    lng = Column(Numeric)


class BomObservation(Base):
    __tablename__ = "bom_observation"

    observation_time = Column(DateTime, primary_key=True)

    station_id = Column(
        Text,
        ForeignKey("bom_station.code", name="fk_bom_observation_station_code"),
        primary_key=True,
    )
    station = relationship("BomStation")
    temp_apparent = Column(Numeric)
    temp_air = Column(Numeric)
    press_qnh = Column(Numeric)
    wind_dir = Column(Text, nullable=True)
    wind_spd = Column(Numeric)



# Copyright 2010-2020 Kurt McKee <contactme@kurtmckee.org>
# Copyright 2002-2008 Mark Pilgrim
# All rights reserved.
#
# This file is a part of feedparser.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 'AS IS'
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import
from __future__ import unicode_literals

import datetime

timezonenames = {
    'ut': 0, 'gmt': 0, 'z': 0,
    'adt': -3, 'ast': -4, 'at': -4,
    'edt': -4, 'est': -5, 'et': -5,
    'cdt': -5, 'cst': -6, 'ct': -6,
    'mdt': -6, 'mst': -7, 'mt': -7,
    'pdt': -7, 'pst': -8, 'pt': -8,
    'a': -1, 'n': 1,
    'm': -12, 'y': 12,
}
# W3 date and time format parser
# http://www.w3.org/TR/NOTE-datetime
# Also supports MSSQL-style datetimes as defined at:
# http://msdn.microsoft.com/en-us/library/ms186724.aspx
# (basically, allow a space as a date/time/timezone separator)


def _parse_date_w3dtf(datestr):
    if not datestr.strip():
        return None
    parts = datestr.lower().split('t')
    if len(parts) == 1:
        # This may be a date only, or may be an MSSQL-style date
        parts = parts[0].split()
        if len(parts) == 1:
            # Treat this as a date only
            parts.append('00:00:00z')
    elif len(parts) > 2:
        return None
    date = parts[0].split('-', 2)
    if not date or len(date[0]) != 4:
        return None
    # Ensure that `date` has 3 elements. Using '1' sets the default
    # month to January and the default day to the 1st of the month.
    date.extend(['1'] * (3 - len(date)))
    try:
        year, month, day = [int(i) for i in date]
    except ValueError:
        # `date` may have more than 3 elements or may contain
        # non-integer strings.
        return None
    if parts[1].endswith('z'):
        parts[1] = parts[1][:-1]
        parts.append('z')
    # Append the numeric timezone offset, if any, to parts.
    # If this is an MSSQL-style date then parts[2] already contains
    # the timezone information, so `append()` will not affect it.
    # Add 1 to each value so that if `find()` returns -1 it will be
    # treated as False.
    loc = parts[1].find('-') + 1 or parts[1].find('+') + 1 or len(parts[1]) + 1
    loc = loc - 1
    parts.append(parts[1][loc:])
    parts[1] = parts[1][:loc]
    time = parts[1].split(':', 2)
    # Ensure that time has 3 elements. Using '0' means that the
    # minutes and seconds, if missing, will default to 0.
    time.extend(['0'] * (3 - len(time)))
    if parts[2][:1] in ('-', '+'):
        try:
            tzhour = int(parts[2][1:3])
            tzmin = int(parts[2][4:])
        except ValueError:
            return None
        if parts[2].startswith('-'):
            tzhour = tzhour * -1
            tzmin = tzmin * -1
    else:
        tzhour = timezonenames.get(parts[2], 0)
        tzmin = 0
    try:
        hour, minute, second = [int(float(i)) for i in time]
    except ValueError:
        return None
    # Create the datetime object and timezone delta objects
    try:
        stamp = datetime.datetime(year, month, day, hour, minute, second)
    except ValueError:
        return None
    delta = datetime.timedelta(0, 0, 0, 0, tzmin, tzhour)
    # Return the date and timestamp in a UTC 9-tuple
    try:
        return (stamp - delta).utctimetuple()
    except (OverflowError, ValueError):
        # IronPython throws ValueErrors instead of OverflowErrors
        return None



import datetime
import os
import re
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import requests

from dms_daemon import CONFIG


def notice_list():
    for root, _, files in os.walk("/data/marble/nemweb/MARKET_NOTICES/"):
        for f in files:
            yield os.path.join(root, f)


def read_notice(fn="/data/marble/nemweb/MARKET_NOTICES/2017-12-21/60425.txt"):
    with open(fn, "r") as f:
        return f.readlines()


def check_notice_list():
    for _, _, files in os.walk("/data/marble/nemweb/MARKET_NOTICES/"):
        for f in files:
            None
            # check_notice(fn=os.path.join(root,f))
            # warning, uncommenting could send lots of slack alerts


def check_notice(
    fn,
    notice_types=[
        "POWER SYSTEM EVENTS",
        "GENERAL NOTICE",
        "RESERVE NOTICE",
        "MARKET INTERVENTION",
    ],
):
    market_notice = read_notice(fn)

    if any([nt in market_notice[11] for nt in notice_types]):
        slack_alert(market_notice)

    if market_notice[14].find("Forecast Lack Of Reserve Level 2") != -1:
        send(market_notice)

    if market_notice[12].find("LRC/LOR1/LOR2/LOR3") != -1:
        try:
            reserve_notice(market_notice)
        except Exception as E:
            r = requests.post(
                "https://hooks.slack.com/services/{0}".format(
                    CONFIG["slack_hooks"]["error_message"]
                ),
                json={"text": E.args[0]},
            )
            raise E


def slack_alert(market_notice):
    mn_text = "".join(market_notice)
    message_type = market_notice[11].split(":")[1].strip()
    message = "*Market Notice: {0}*\n\n{1}".format(message_type, mn_text)

    requests.post(
        "https://hooks.slack.com/services/{0}".format(
            CONFIG["slack_hooks"]["market_notice"]
        ),
        json={"text": message},
    )


def reserve_notice(market_notice):
    lor_pattern = re.compile("LOR([0-9]{1})")
    lor_match = re.search(lor_pattern, market_notice[14])
    if lor_match == None:
        if "RESERVE NOTICE" in market_notice[14]:
            None
        else:
            raise Exception("No LOR in external reference")
    else:
        lor_notice(market_notice, lor_match)


def lor_notice(market_notice, lor_match):
    region_pattern = re.compile(r"the (.{2,15}) \w{1}egion")
    region_match = re.search(region_pattern, market_notice[14])
    if region_match == None:
        raise Exception("No region in external reference")

    lor_category_pattern = re.compile(r"(\w{1}orecast|\w{1}ctual|\w{1}ancellation)")
    lor_category_match = re.search(lor_category_pattern, market_notice[14])
    if lor_category_match == None:
        raise Exception("No LOR type (not Forecast or Actual)")
    elif lor_category_match.group(1).lower() == "cancellation":
        lor_cancellation_template(
            int(lor_match.group(1)), region_match.group(1), lor_category_match.group(1)
        )
    else:
        lor_update(market_notice, lor_match, region_match, lor_category_match)
    return region_match, lor_category_match


def lor_update(market_notice, lor_match, region_match, lor_category_match):

    if market_notice[14].find("Update") != -1:
        lor_type = "New"
    else:
        lor_type = "Update"

    from_time2 = None
    to_time2 = None

    try:
        from_time = time_parse(string="".join(market_notice), start="(F|f)rom")
        to_time = time_parse(string="".join(market_notice), start="(T|t)o")
    except:
        try:
            from_time, to_time = time_parse2(string="".join(market_notice))
        except:
            try:
                from_time, to_time = time_parse3(string="".join(market_notice))
            except:
                try:
                    from_time, to_time, from_time2, to_time2 = multi_time_parse(
                        string="".join(market_notice)
                    )
                except:
                    raise Exception("Bad datestring format")

    required = reserve_required(market_notice)
    available = reserve_available(market_notice)

    lor_template(
        int(lor_match.group(1)),
        lor_category_match.group(1),
        region_match.group(1),
        lor_type,
        from_time,
        to_time,
        int(required),
        int(available),
        from_dt2=from_time2,
        to_dt2=to_time2,
    )


def reserve_available(market_notice):
    reserve_available_pattern = re.compile(r"reserve available is (\d{3,4}) MW")
    reserve_available_match = re.search(reserve_available_pattern, "".join(market_notice))
    if reserve_available_match == None:
        raise Exception("Bad reserve requirement format")
    return reserve_available_match.group(1)


def reserve_required(market_notice):
    reserve_requirement_pattern = re.compile(
        r"reserve requirement is (\d{3,4}|\d{1},\d{3}) MW"
    )
    reserve_requirement_match = re.search(
        reserve_requirement_pattern, "".join(market_notice)
    )
    if reserve_requirement_match == None:
        raise Exception("Bad reserve requirement format")
    return reserve_requirement_match.group(1).replace(",", "")


def lor_template(
    level,
    category,
    region,
    lor_type,
    from_dt,
    to_dt,
    reserve_req,
    reserve_avail,
    from_dt2=None,
    to_dt2=None,
):
    colors = {1: "#ffc100", 2: "#ff7400", 3: "ff0000"}
    title = "LOR{0} {2} for {1} ({3})".format(
        level, region, category.lower(), lor_type.lower()
    )
    details = "({0}MW shortfall between {1} and {2})".format(
        reserve_req - reserve_avail, from_dt, to_dt
    )
    if from_dt2 != None:
        details = "({0}MW shortfall between {1} to {2}, and {3} to {4})".format(
            reserve_req - reserve_avail, from_dt, to_dt, from_dt2, to_dt2
        )
    r = requests.post(
        "https://hooks.slack.com/services/{0}".format(
            CONFIG["slack_hooks"]["reserve_notice"]
        ),
        json={
            "attachments": [{"title": title, "color": colors[level]}, {"text": details}]
        },
    )


def lor_cancellation_template(level, region, category):
    colors = {1: "#36a64f", 2: "#ffc100", 3: "#ff7400"}
    title = "LOR{0} {2} for {1}".format(level, region, category.lower())
    requests.post(
        "https://hooks.slack.com/services/{0}".format(
            CONFIG["slack_hooks"]["reserve_notice"]
        ),
        json={"attachments": [{"title": title, "color": colors[level]}]},
    )


def time_parse(string, start="(F|f)rom"):
    time_pattern = re.compile(
        r"{0} (\d{{2}})(\d{{2}}) hrs (\d{{2}})/(\d{{2}})/(\d{{4}})".format(start)
    )
    time_match = re.search(time_pattern, string)
    _, H, M, d, m, y = time_match.groups()
    return datetime.datetime(int(y), int(m), int(d), int(H), int(M))


def time_parse2(string):
    time_pattern = re.compile(
        r"(F|f)rom (\d{2})(\d{2}) hrs (T|t)o (\d{2})(\d{2}) hrs (\d{2})/(\d{2})/(\d{4})"
    )
    time_match = re.search(time_pattern, string)
    _, H1, M1, _, H2, M2, d, m, y = time_match.groups()
    return (
        datetime.datetime(int(y), int(m), int(d), int(H1), int(M1)),
        datetime.datetime(int(y), int(m), int(d), int(H2), int(M2)),
    )


def time_parse3(string):
    time_pattern = re.compile(
        r"(F|f)rom (\d{2})(\d{2}) hrs (T|t)o (\d{2})(\d{2}) hrs on (\d{2})/(\d{2})/(\d{4})"
    )
    time_match = re.search(time_pattern, string)
    _, H1, M1, _, H2, M2, d, m, y = time_match.groups()
    return (
        datetime.datetime(int(y), int(m), int(d), int(H1), int(M1)),
        datetime.datetime(int(y), int(m), int(d), int(H2), int(M2)),
    )


def multi_time_parse(string):
    time_pattern = re.compile(
        r"(F|f)rom (\d{2})(\d{2}) hrs (T|t)o (\d{2})(\d{2}) hrs and (\d{2})(\d{2}) hrs (T|t)o (\d{2})(\d{2}) hrs (\d{2})/(\d{2})/(\d{4})"
    )
    time_match = re.search(time_pattern, string)
    _, H1, M1, _, H2, M2, H3, M3, _, H4, M4, d, m, y = time_match.groups()
    return (
        datetime.datetime(int(y), int(m), int(d), int(H1), int(M1)),
        datetime.datetime(int(y), int(m), int(d), int(H2), int(M2)),
        datetime.datetime(int(y), int(m), int(d), int(H3), int(M3)),
        datetime.datetime(int(y), int(m), int(d), int(H4), int(M4)),
    )


def login():
    server = smtplib.SMTP("smtp.gmail.com:587")
    server.ehlo()
    server.starttls()
    server.login(CONFIG["gmail"]["username"], CONFIG["gmail"]["password"])
    return server


def create_msg(market_notice):
    mn_text = "".join(market_notice)
    message_type = market_notice[11].split(":")[1].strip()

    msg = MIMEMultipart()
    msg["From"] = CONFIG["gmail"]["username"]
    msg["To"] = CONFIG["gmail"]["send_to"]
    msg["Subject"] = message_type
    msg.add_header("Content-Type", "text")

    msg.attach(MIMEText(mn_text, "plain"))
    return msg


def send(market_notice):
    server = login()
    msg = create_msg(market_notice)
    server.sendmail(
        CONFIG["gmail"]["username"], CONFIG["gmail"]["send_to"], msg.as_string()
    )
    server.close()


test_file = "/data/marble/nemweb/MARKET_NOTICES/2017-10-24/59591.txt"

line_12 = [
    "Details of Non-conformance/Conformance",
    "Emergency events/conditions",
    "Inter-Regional Transfer limit variation",
    "LRC/LOR1/LOR2/LOR3",
    "New/Modified Constraints",
    "Prices are not firm due to an MII or OCD event",
    "Prices have been reviewed and remain unchanged",
    "Reclassify contingency events",
    "Reserve Contract / Direction / Instruction",
    "Status of Market Systems",
    "Subjects not covered in specific notices",
]

line_11 = [
    "CONSTRAINTS",
    "GENERAL NOTICE",
    "INTER-REGIONAL TRANSFER",
    "MARKET INTERVENTION",
    "MARKET SYSTEMS",
    "NON-CONFORMANCE",
    "POWER SYSTEM EVENTS",
    "PRICES SUBJECT TO REVIEW",
    "PRICES UNCHANGED",
    "RECLASSIFY CONTINGENCY",
    "RESERVE NOTICE",
]



import csv

from geojson import Feature, FeatureCollection, Point, dumps
from opennem.controllers.stations import get_stations
from opennem.core.facility_duid_map import duid_is_retired
from opennem.db import get_database_session
from opennem.exporter.encoders import OpenNEMGeoJSONEncoder

__all__ = ["stations_geojson_serialize"]


def stations_geojson_records():
    db = get_database_session()

    stations = get_stations(db)

    records = []

    for station in stations:

        geom = None

        if station.lat and station.lng:
            geom = Point((station.lat, station.lng))

        f = Feature(geometry=geom)

        f.properties = {
            "oid": station.oid,
            "ocode": station.ocode,
            "station_id": station.id,
            "station_code": station.network_code,
            "facility_id": station.network_code,
            "network": station.network.label,
            "network_country": station.network.country,
            "state": station.state.upper() if station.state else None,
            "postcode": station.postcode,
            "name": station.name,
            "capacity_registered": station.capacity_registered,
            "capacity_aggregate": station.capacity_aggregate,
            "duid_data": [],
        }

        for facility in station.facilities:
            if facility.fueltech_id is None:
                continue

            if facility.status_id is None:
                continue

            if duid_is_retired(facility.code):
                continue

            if facility.active == False:
                continue

            f.properties["duid_data"].append(
                {
                    "oid": facility.oid,
                    "duid": facility.duid,
                    "fuel_tech": facility.fueltech_id,
                    "fuel_tech_label": facility.fueltech_label,
                    "fuel_tech_renewable": facility.fueltech.renewable
                    if facility.fueltech
                    else None,
                    "commissioned_date": facility.registered,
                    "decommissioned_date": facility.deregistered,
                    "status": facility.status_id,
                    "status_label": facility.status_label,
                    "unit_id": facility.unit_id,
                    "unit_number": facility.unit_number,
                    "unit_size": facility.unit_capacity,
                    "unit_alias": facility.unit_alias,
                    # capacities for the unit
                    "capacity_registered": facility.capacity_registered,
                    "capacity_aggregate": facility.capacity_aggregate,
                    # network specific fields (DUID is one)
                    "network_region": facility.network_region,
                }
            )

        if len(f.properties["duid_data"]) > 0:
            records.append(f)

    return records


def stations_geojson_serialize():
    crs = {
        "type": "name",
        "properties": {"name": "urn:ogc:def:crs:OGC:1.3:CRS84"},
    }

    stations = stations_geojson_records()

    geoj = FeatureCollection(stations, crs=crs, name="opennem")

    geoj["name"] = "opennem_stations"
    geoj["crs"] = crs

    geoj_string = dumps(geoj, indent=4, cls=OpenNEMGeoJSONEncoder)

    return geoj_string


if __name__ == "__main__":
    stations_geojson_serialize()



from typing import Optional

from pydantic import BaseModel, ValidationError, validator

from opennem.core.normalizers import (
    clean_capacity,
    is_number,
    normalize_duid,
    normalize_string,
    participant_name_filter,
    station_name_cleaner,
)

from .opennem import OpennemBaseSchema


class OpennemParticipant(OpennemBaseSchema):
    code: Optional[str]
    name: str
    country: str = "au"
    abn: Optional[str]

    @classmethod
    @validator("code")
    def code_clean(cls, v):
        return normalize_duid(v)

    @classmethod
    @validator("name")
    def name_clean(cls, v):
        name = participant_name_filter(v)

        return name



import json
import logging
import os
from datetime import date, datetime
from decimal import Decimal

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import MultipleResultsFound

from opennem.core.facilitystations import facility_station_join_by_name
from opennem.core.facilitystatus import map_v3_states
from opennem.core.normalizers import (
    clean_capacity,
    name_normalizer,
    normalize_duid,
    station_name_cleaner,
)
from opennem.core.station_duid_map import (
    facility_has_station_remap,
    facility_map_station,
)
from opennem.db import db_connect
from opennem.db.load_fixtures import load_fixture
from opennem.db.models.opennem import Facility, Station

logger = logging.getLogger(__name__)


def fueltech_map(fueltech):
    if fueltech == "brown_coal":
        return "coal_brown"

    if fueltech == "black_coal":
        return "coal_black"

    if fueltech == "solar":
        return "solar_utility"

    if fueltech == "biomass":
        return "bioenergy_biomass"

    return fueltech


def map_network_region(network_region: str) -> str:
    """
        Map network regions from old to new

        Note that network regions aren't really geos
        and there are networks within geos like DKIS (NT)
        and NWIS (WA) that need to retain their network_region
    """
    if not network_region or not type(network_region) is str:
        return network_region

    network_region = network_region.strip()

    if network_region == "WA1":
        return "WEM"

    return network_region


def load_opennem_facilities():
    station_fixture = load_fixture("facility_registry.json")

    stations = [{"station_code": k, **v} for k, v in station_fixture.items()]

    engine = db_connect()
    session = sessionmaker(bind=engine)
    s = session()

    for station_data in stations:
        station = None

        facilities = [
            {"code": k, **v} for k, v in station_data["duid_data"].items()
        ]

        # update facilities
        for facility_data in facilities:
            facility_duid = facility_data["code"]
            station_name = station_name_cleaner(station_data["display_name"])
            # station_code = normalize_duid(station_data["station_code"])
            station_code = facility_map_station(
                facility_duid, normalize_duid(station_data["station_code"])
            )
            station_state = map_v3_states(station_data["status"]["state"])
            station_network = (
                "WEM" if station_data["location"]["state"] == "WA" else "NEM"
            )

            station = (
                s.query(Station)
                .filter(Station.network_code == station_code)
                .one_or_none()
            )

            if not station:
                station = Station(
                    network_id=station_network,
                    code=station_code,
                    network_code=station_code,
                    name=station_name,
                    network_name=station_data["display_name"],
                    created_by="opennem.load_facilities",
                )
                logger.info(
                    "Created station: {} {} ".format(
                        station_name, station_code
                    )
                )

                s.add(station)
                s.commit()

            facility_status = station_state
            facility_network_region = map_network_region(
                station_data["region_id"]
            )
            facility_fueltech = (
                fueltech_map(facility_data["fuel_tech"])
                if "fuel_tech" in facility_data
                else None
            )
            facility_capacity = (
                clean_capacity(facility_data["registered_capacity"])
                if "registered_capacity" in facility_data
                else None
            )

            facility = None

            try:
                facility = (
                    s.query(Facility)
                    .filter(Facility.network_code == facility_duid)
                    .one_or_none()
                )
            except MultipleResultsFound:
                logger.error(
                    "Multiple facilities found for duid {}".format(
                        facility_duid
                    )
                )

                # facility = (
                #     s.query(Facility)
                #     .filter(Facility.network_code == facility_duid)
                #     .first()
                # )
                continue

            if not facility:
                facility = Facility(
                    code=facility_duid,
                    network_code=facility_duid,
                    network_region=facility_network_region,
                    created_by="opennem.load_facilities",
                )

                logger.info(
                    "Created facility: {} {} to station {} ".format(
                        facility_duid, facility_fueltech, station_code
                    )
                )

            if not facility.unit_id:
                facility.unit_id = 1
                facility.unit_number = 1

            if facility_capacity and not facility.unit_capacity:
                facility.unit_capacity = facility_capacity

            if facility_capacity and not facility.capacity_registered:
                facility.capacity_registered = facility_capacity

            if facility_fueltech and not facility.fueltech_id:
                facility.fueltech_id = facility_fueltech

            if not facility.status_id:
                facility.status_id = facility_status

            if not facility.station:
                facility.station = station

            s.add(facility)
            s.commit()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    load_opennem_facilities()



import csv
import json
import logging
import os
from datetime import date, datetime
from decimal import Decimal

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker

from opennem.core.facilitystations import facility_station_join_by_name
from opennem.core.facilitystatus import map_v3_states
from opennem.core.normalizers import (
    clean_capacity,
    name_normalizer,
    normalize_duid,
    station_name_cleaner,
)
from opennem.db import db_connect
from opennem.db.models.bom import BomStation
from opennem.db.models.opennem import (
    Facility,
    FacilityStatus,
    FuelTech,
    Network,
    Participant,
    Station,
)

FIXTURE_PATH = os.path.join(os.path.dirname(__file__), "fixtures")

engine = db_connect()
session = sessionmaker(bind=engine)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("opennem.fixtureloader")
logger.setLevel(logging.INFO)


def fueltech_map(fueltech):
    if fueltech == "brown_coal":
        return "coal_brown"

    if fueltech == "black_coal":
        return "coal_black"

    if fueltech == "solar":
        return "solar_utility"

    if fueltech == "biomass":
        return "bioenergy_biomass"

    return fueltech


def load_fixture(fixture_name):
    fixture_path = os.path.join(FIXTURE_PATH, fixture_name)

    if not os.path.isfile(fixture_path):
        raise Exception("Not a file: {}".format(fixture_path))

    fixture = None

    with open(fixture_path) as fh:
        fixture = json.load(fh)

    return fixture


def load_fueltechs():
    fixture = load_fixture("fueltechs.json")

    s = session()

    for fueltech in fixture:
        ft = FuelTech(
            created_by="opennem.fixture",
            code=fueltech["code"],
            label=fueltech["label"],
            renewable=fueltech["renewable"],
        )

        try:
            s.add(ft)
            s.commit()
        except Exception:
            logger.error("Have {}".format(ft.code))


def load_facilitystatus():
    fixture = load_fixture("facility_status.json")

    s = session()

    for status in fixture:
        ft = FacilityStatus(
            created_by="opennem.fixture",
            code=status["code"],
            label=status["label"],
        )

        try:
            s.add(ft)
            s.commit()
        except Exception:
            logger.error("Have {}".format(ft.code))


def load_networks():
    fixture = load_fixture("networks.json")

    s = session()

    for network in fixture:
        ft = Network(
            created_by="opennem.fixture",
            code=network["code"],
            label=network["label"],
            country=network["country"],
        )

        try:
            s.add(ft)
            s.commit()
        except Exception:
            logger.error("Have {}".format(ft.code))


def parse_date(date_str):
    dt = datetime.strptime(date_str, "%Y%m%d")
    return dt.date()


def parse_fixed_line(line):
    return (
        str(line[:6]),
        str(line[8:11]).strip(),
        str(line[18:59]).strip(),
        parse_date(line[59:67]),
        Decimal(line[75:83]),
        Decimal(line[84:]),
    )


def load_bom_stations():
    s = session()

    with open(os.path.join(FIXTURE_PATH, "stations_db.txt")) as fh:
        lines = fh.readlines()
        for line in lines:
            code, state, name, registered, lat, lng = parse_fixed_line(line)
            station = BomStation(
                code=code,
                state=state,
                name=name,
                registered=registered,
                lat=lat,
                lng=lng,
            )

            try:
                s.add(station)
                s.commit()
            except Exception:
                logger.error("Have {}".format(station.code))


def update_existing_geos():
    station_fixture = load_fixture("facility_registry.json")

    stations = [{"station_code": k, **v} for k, v in station_fixture.items()]

    s = session()

    for station_data in stations:
        station = None

        station_name = station_name_cleaner(station_data["display_name"])
        station_code = normalize_duid(station_data["station_code"])
        station_state = map_v3_states(station_data["status"]["state"])

        station = (
            s.query(Station)
            .filter(Station.network_code == station_code)
            .one_or_none()
        )

        if not station:
            logger.info("Could not find station {}".format(station_code))
            continue

        if (
            "location" in station_data
            and "latitude" in station_data["location"]
            and station_data["location"]["latitude"]
        ):
            station.geom = (
                "SRID=4326;POINT({} {})".format(
                    station_data["location"]["latitude"],
                    station_data["location"]["longitude"],
                ),
            )
            station.geocode_processed_at = datetime.now()
            station.geocode_by = "opennem"
            station.geocode_approved = True

            station.updated_by = "fixture.registry"

        s.add(station)

        logger.info(
            "Updated station geo location {} ({})".format(
                station.code, station.name,
            )
        )

        facilities = [
            {"code": k, **v} for k, v in stations[0]["duid_data"].items()
        ]

        # update fueltechs
        for facility_data in facilities:
            facility_duid = facility_data["code"]
            facility_fueltech = fueltech_map(facility_data["fuel_tech"])

            facility = (
                s.query(Facility)
                .filter(Facility.network_code == facility_duid)
                .first()
            )

            if not facility:
                logger.error(
                    "Could not find existing facility {} for station {}".format(
                        facility_duid, station_code
                    )
                )
                continue

            if not facility.fueltech_id:
                facility.fueltech_id = facility_fueltech

            if facility.fueltech_id != facility_fueltech:
                logger.error(
                    "Fueltech mismatch for {}. Old is {} and new is {}".format(
                        station_code, facility_fueltech, station.fueltech_id
                    )
                )

            s.add(facility)

        s.commit()


if __name__ == "__main__":
    load_fueltechs()
    load_facilitystatus()
    load_networks()
    load_bom_stations()



from opennem.db import db_connect
from opennem.db.models.bom import metadata as metadata_bom
from opennem.db.models.opennem import metadata as metadata_opennem

if __name__ == "__main__":
    engine = db_connect()

    metadata_opennem.drop_all(engine)
    metadata_opennem.create_all(engine)
    print("Created opennem schema.")

    metadata_bom.drop_all(engine)
    metadata_bom.create_all(engine)
    print("Created bom schema ....")

    print("Done initdb.")



from opennem.pipelines.wem.participant import (
    WemStoreLiveParticipant,
    WemStoreParticipant,
)
from opennem.spiders.wem import WemCurrentSpider


class WemParticipantSpider(WemCurrentSpider):
    name = "au.wem.participant"

    pipelines_extra = set([WemStoreParticipant,])

    start_url = "http://data.wa.aemo.com.au/public/public-data/datafiles/participants/participants.csv"


class WemParticipantLiveSpider(WemCurrentSpider):
    name = "au.wem.live.participant"

    pipelines_extra = set([WemStoreLiveParticipant,])

    start_url = "https://aemo.com.au/aemo/data/wa/infographic/participant.csv"



import json
from datetime import datetime
from typing import List, Optional, Union

from geoalchemy2.elements import WKBElement
from pydantic import BaseModel

from opennem.core.dispatch_type import DispatchType
from opennem.exporter.encoders import OpenNEMJSONEncoder


class OpennemBaseSchema(BaseModel):

    created_by: Optional[str]
    updated_by: Optional[str]

    processed_at: Optional[datetime] = None
    created_at: datetime
    updated_at: Optional[datetime] = None

    class Config:
        orm_mode = True
        anystr_strip_whitespace = True

        #
        arbitrary_types_allowed = True
        # validate_assignment = True

        json_encoders = {}

        def json_dumps(self):
            print("json", self.__values__)
            return json.dumps(self.__values__, cls=OpenNEMJSONEncoder,)


class FueltechSchema(OpennemBaseSchema):
    code: str
    label: str
    renewable: bool


class NetworkSchema(OpennemBaseSchema):
    code: str
    country: str
    label: str


class FacilityStatusSchema(OpennemBaseSchema):
    code: str
    label: str


class ParticipantSchema(OpennemBaseSchema):
    id: int
    code: Optional[str]
    name: str
    network_name: Optional[str] = None
    network_code: Optional[str] = None
    country: Optional[str] = None
    abn: Optional[str] = None


class StationBaseSchema(OpennemBaseSchema):
    id: int


class FacilityBaseSchema(OpennemBaseSchema):
    id: int


class FacilitySchema(FacilityBaseSchema):

    participant: ParticipantSchema
    participant_id: Optional[str]

    fueltech: FueltechSchema
    fueltech_id: Optional[str]

    status: FacilityStatusSchema
    status_id: Optional[str]

    station: StationBaseSchema
    station_id: int

    code: str

    network_code: str
    network_region: Optional[str]
    network_name: Optional[str]

    active: bool = True

    dispatch_type: DispatchType

    capacity_registered: Optional[float]

    registered: Optional[datetime]
    deregistered: Optional[datetime]

    unit_id: Optional[int]
    unit_number: Optional[int]
    unit_alias: Optional[str]
    unit_capacity: Optional[float]

    # virtual methods
    capacity_aggregate: Optional[int]
    duid: str
    oid: str
    ocode: str
    status_label: Optional[str]
    fueltech_label: Optional[str]


class StationSchema(StationBaseSchema):
    network: NetworkSchema
    network_id: str

    participant: Optional[ParticipantSchema] = None
    participant_id: Optional[str]

    # facilities: List[FacilitySchema] = []

    code: Optional[str] = None
    name: Optional[str] = None

    address1: Optional[str] = ""
    address2: Optional[str] = ""
    locality: Optional[str] = ""
    state: Optional[str] = ""
    postcode: Optional[str] = ""

    # Original network fields
    network_code: Optional[str]
    network_name: Optional[str]

    # Geo fields
    # place_id: Optional[str]
    geocode_approved: bool = False
    geocode_skip: bool = False
    geocode_processed_at: Optional[datetime] = None
    geocode_by: Optional[str]
    # geom: Optional[WKBElement] = None
    # boundary: Optional[WKBElement] = None

    # virtual methods
    capacity_aggregate: Optional[int]
    oid: str
    ocode: str
    lat: Optional[float]
    lng: Optional[float]


class StationSubmission(BaseModel):
    name: str
    network_id: str

    class Config:
        orm_mode = True



import csv
import logging
from datetime import datetime, timedelta

from scrapy.exceptions import DropItem
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import text

from opennem.db.models.opennem import FacilityScada
from opennem.pipelines import DatabaseStoreBase
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class WemStoreFacilityScada(DatabaseStoreBase):
    def parse_interval(self, date_str):
        return datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        csvreader = csv.DictReader(item["content"].split("\n"))

        # Get all facility codes
        #  It's faster to do this and bulk insert than it is to
        #  catch integrityerrors

        q = self.engine.execute(text("select code from wem_facility"))

        all_facilities = list(set([i[0] for i in q.fetchall()]))

        objects = []
        keys = []

        for row in csvreader:
            if not row["Facility Code"] in all_facilities:
                continue

            row_key = "{}_{}".format(
                row["Trading Interval"], row["Facility Code"]
            )

            item = FacilityScada(
                trading_interval=self.parse_interval(row["Trading Interval"]),
                facility_id=row["Facility Code"],
                eoi_quantity=row["EOI Quantity (MW)"] or None,
                generated=row["Energy Generated (MWh)"] or None,
            )

            if row_key not in keys:
                objects.append(item)
                keys.append(row_key)

        try:
            s.bulk_save_objects(objects)
            s.commit()
        except Exception as e:
            logger.error("Error: {}".format(e))
        finally:
            s.close()

        return len(objects)


class WemStoreLiveFacilityScada(DatabaseStoreBase):
    """
        Store live facility scada data.

    """

    def parse_interval(self, date_str):
        return datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")

        # try:
        #     return datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")
        # except Exception:
        #     pass

        # return datetime.strptime(date_str, "%d/%m/%Y %H:%M:%S %p")

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        csvreader = csv.DictReader(item["content"].split("\n"))

        # Get all facility codes

        q = self.engine.execute(text("select code from wem_facility"))

        all_facilities = list(set([i[0] for i in q.fetchall()]))

        objects = []
        keys = []
        last_asat = None

        for row in csvreader:
            if not row["FACILITY_CODE"] in all_facilities:
                logger.error(
                    "Do not have facility: {}".format(row["FACILITY_CODE"])
                )
                continue

            if row["AS_AT"] != "":
                last_asat = self.parse_interval(row["AS_AT"])

            if not last_asat or type(last_asat) is not datetime:
                logger.error("Invalid row or no datetime")
                continue

            # We need to pivot the table since columns are time intervals
            for i in range(1, 48):

                column = f"I{i:02}"

                if column not in row:
                    logger.error(
                        "Do not have data for interval {}".format(column)
                    )
                    continue

                if i > 0:
                    interval = last_asat - timedelta(minutes=(i - 1) * 30)
                else:
                    interval = last_asat

                row_key = "{}_{}".format(interval, row["FACILITY_CODE"])

                val = None

                try:
                    val = float(row[column]) / 2 or None
                except ValueError:
                    pass

                item = FacilityScada(
                    trading_interval=interval,
                    facility_id=row["FACILITY_CODE"],
                    generated=val,
                )

                if row_key not in keys:
                    # objects.append(item)
                    s.add(item)
                    # s.save()
                    keys.append(row_key)

        try:
            # s.bulk_save_objects(objects)
            s.commit()
        except Exception as e:
            logger.error("Error: {}".format(e))
        finally:
            s.close()

        return len(objects)



import logging
import os

from smart_open import open

logger = logging.getLogger(__name__)

S3_EXPORT_DEFAULT_BUCKET = "s3://data.opennem.org.au/v3/"

UPLOAD_ARGS = {
    "ContentType": "application/json",
}


def write_to_s3(file_path: str, data: str) -> int:
    """
        Write data to an s3 path
    """
    s3_save_path = os.path.join(S3_EXPORT_DEFAULT_BUCKET, file_path)
    write_count = 0

    with open(
        s3_save_path,
        "w",
        transport_params=dict(multipart_upload_kwargs=UPLOAD_ARGS),
    ) as fh:
        write_count = fh.write(data)

    logger.info("Wrote {} to {}".format(len(data), s3_save_path))

    return write_count



import csv
from io import StringIO

from opennem.controllers.stations import get_stations
from opennem.core.facility_duid_map import duid_is_retired
from opennem.db import get_database_session


def stations_csv_records():
    db = get_database_session()
    stations = get_stations(db)

    records = []

    for station in stations:
        for facility in station.facilities:
            if facility.fueltech_id is None:
                continue

            if facility.status_id is None:
                continue

            if duid_is_retired(facility.code):
                continue

            if facility.active == False:
                continue

            rec = {
                "name": station.name,
                # "oid": station.oid,
                "ocode": station.ocode,
                "code": facility.duid,
                "region": facility.network_region,
                "status": facility.status_id,
                "fueltech": facility.fueltech_id,
                "unit_id": facility.unit_id,
                "unit_num": facility.unit_number,
                "unit_cap": facility.capacity_aggregate,
                "station_cap_agg": station.capacity_aggregate,
                "station_cap_registered": station.capacity_registered,
                "added_by": facility.created_by,
                "updated_by": facility.updated_by,
            }
            records.append(rec)

    return records


def stations_csv_serialize(csv_stream=None):

    if not csv_stream:
        csv_stream = StringIO()

    csv_records = stations_csv_records()

    csv_fieldnames = csv_records[0].keys()

    csvwriter = csv.DictWriter(csv_stream, fieldnames=csv_fieldnames)
    csvwriter.writeheader()
    csvwriter.writerows(csv_records)

    return csv_stream



import logging

logger = logging.getLogger(__name__)


def map_v3_states(state):
    """
        Maps a current v1 version opennem facility_registry.json state
        to a v3 state

    """
    state = state.lower().strip()

    if state == "commissioned":
        return "operating"

    if state == "decommissioned":
        return "retired"

    return state


def lookup_facility_status(unit_status):

    unit_status = unit_status or ""

    unit_status = unit_status.lower().strip()

    if unit_status.startswith("in service"):
        return "operating"

    if unit_status.startswith("in commissioning"):
        return "commissioning"

    if unit_status.startswith("committed"):
        return "committed"

    if unit_status.startswith("maturing"):
        return "maturing"

    if unit_status.startswith("emerging"):
        return "emerging"

    logger.error(
        "Could not find status for unit status: {}".format(unit_status)
    )

    return None



import csv

import scrapy

from opennem.pipelines.wem.facility_scada import (
    WemStoreFacilityScada,
    WemStoreLiveFacilityScada,
)
from opennem.spiders.wem import WemCurrentSpider, WemHistoricSpider


class WemLiveFacilityScada(WemCurrentSpider):
    name = "au.wem.live.facility_scada"
    pipelines_extra = set([WemStoreLiveFacilityScada])

    start_url = "https://aemo.com.au/aemo/data/wa/infographic/generation.csv"


class WemCurrentFacilityScada(WemCurrentSpider):
    name = "au.wem.current.facility_scada"
    start_url = "http://data.wa.aemo.com.au/public/public-data/datafiles/facility-scada/facility-scada-{year}-{month}.csv"

    pipelines_extra = set([WemStoreFacilityScada])


class WemHistoricFacilityScada(WemHistoricSpider):
    name = "au.wem.historic.facility_scada"
    start_url = "http://data.wa.aemo.com.au/datafiles/facility-scada/"

    pipelines_extra = set([WemStoreFacilityScada])



"""
    Module that generated opennem ids

"""

from typing import Optional

from opennem.core import base24
from opennem.core.normalizers import is_number

# from opennem.db.models.opennem import Station


def get_oid(model) -> str:
    _id = model.id

    if not model.id:
        return None

    parent_oid = None

    if hasattr(model, "station"):
        parent_oid = get_oid(model.station)

    item_oid = base24.dumps(_id)

    oid_components = [
        parent_oid,
        item_oid,
    ]

    oid_string = "_".join([str(i) for i in oid_components if i])

    return oid_string


def oid_to_id(oid: str) -> int:
    pass


def get_network_region(network_region: str) -> Optional[str]:
    """
        Trim the numbers off the end of nem regions

    """
    if (
        network_region
        and type(network_region) is str
        and len(network_region) > 1
    ):
        if is_number(network_region[-1:]):
            return network_region[:-1]

        if network_region.strip().upper() == "WEM":
            return None

        return network_region

    return None


def get_station_facility(station):
    """
        Get the first facility for a station

    """
    if (station.facilities) and len(station.facilities) > 0:
        return station.facilities[0]

    return None


def get_station_code(station):
    if station.code and type(station.code) is str:
        return station.code.replace("_", "")

    if station.network_code and type(station.network_code) is str:
        return station.network_code.replace("_", "")

    return station.id


def get_ocode(station) -> str:
    """

        format: {country}_{network}_{region}_{station}_{id}
    """
    facility = get_station_facility(station)

    values = [
        station.network.country or "au",
        station.network.code or None,
        get_network_region(facility.network_region) or station.state,
        get_station_code(station),
    ]

    ocode_values = []

    [
        ocode_values.append(str(i).lower())
        for i in values
        if i is not None and str(i).lower() not in ocode_values
    ]

    ocode = ".".join(ocode_values)

    return ocode



import json
import os

DATA_PATH = os.path.join(os.path.dirname(__file__), "data")


def load_facility_stations(fixture_name):
    fixture_path = os.path.join(DATA_PATH, fixture_name)

    if not os.path.isfile(fixture_path):
        raise Exception("Not a file: {}".format(fixture_path))

    fixture = None

    with open(fixture_path) as fh:
        fixture = json.load(fh)

    return fixture


STATION_NAME_MAP = load_facility_stations("station_name_maps.json")


def station_map_name(station_name):
    """
        Clean names manually using a map

    """
    if type(station_name) is not str:
        return station_name

    if not type(STATION_NAME_MAP) is dict:
        raise Exception("Error loading station name maps: not a dict")

    if station_name in STATION_NAME_MAP:
        return STATION_NAME_MAP[station_name]

    for map_name in STATION_NAME_MAP.keys():
        if station_name.lower().startswith(map_name.lower()):
            return STATION_NAME_MAP[map_name]

    return station_name



import csv
import logging
from datetime import datetime

from scrapy.exceptions import DropItem
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import text

from opennem.db.models.opennem import BalancingSummary
from opennem.pipelines import DatabaseStoreBase
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class WemStorePulse(DatabaseStoreBase):
    def parse_interval(self, date_str):
        return datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        csvreader = csv.DictReader(item["content"].split("\n"))

        objects = [
            BalancingSummary(
                trading_interval=self.parse_interval(
                    row["TRADING_DAY_INTERVAL"]
                ),
                forecast_load=row["FORECAST_EOI_MW"],
                # generation_scheduled=row["Scheduled Generation (MW)"],
                # generation_non_scheduled=row["Non-Scheduled Generation (MW)"],
                # generation_total=row["Total Generation (MW)"],
                price=row["PRICE"],
            )
            for row in csvreader
        ]

        try:
            s.bulk_save_objects(objects)
            s.commit()
        except Exception as e:
            logger.error("Error: {}".format(e))
            raise e
        finally:
            s.close()

        return len(objects)



"""
    Parse current and v3 facilities registries and take a diff.


"""

import csv
import json
import logging
from datetime import timedelta
from operator import itemgetter
from pprint import pprint

from opennem.utils import logging

logger = logging.getLogger("opennem.diff")
logger.setLevel(logging.DEBUG)

logging.basicConfig(level=logging.INFO)


def normalize_states(state):
    state = state.lower()

    if state == "commissioned":
        return "operating"

    if state == "decommissioned":
        return "retired"

    return state


def normalize_regions(region):
    if region == "WA":
        return "WA1"

    return region


def run_diff():
    logger.info("Running facility diff")

    with open("opennem/db/fixtures/facility_registry.json") as fh:
        fac_current = json.load(fh)

    remapped = []

    for k, v in fac_current.items():
        for duid, fac in v["duid_data"].items():
            i = [
                v["display_name"] or "",
                k,
                normalize_regions(v["region_id"]) or "",
                normalize_states(v["status"]["state"]),
                duid or "",
                fac["fuel_tech"] if "fuel_tech" in fac else "",
                fac["registered_capacity"]
                if "registered_capacity" in fac
                else 0,
            ]
            remapped.append(i)

    remapped = sorted(remapped, key=itemgetter(2, 0, 4, 6))

    with open("data/facility_diff/facilities_current.csv", "w") as fh:
        csvwriter = csv.writer(fh)
        for line in remapped:
            csvwriter.writerow(line)

    with open("data/facility_diff/au_facilities.json") as fh:
        fac_v3 = json.load(fh)

    fac_v3 = fac_v3["features"]
    remapped = []

    for f in fac_v3:
        for fac in f["properties"]["duid_data"]:
            i = [
                f["properties"]["name"] or "",
                # fac["duid"],
                f["properties"]["oid"],
                f["properties"]["station_code"] or "",
                normalize_regions(f["properties"]["network_region"]) or "",
                fac["status"].lower(),
                fac["duid"] or "",
                fac["fuel_tech"] if "fuel_tech" in fac else "",
                fac["registered_capacity"]
                if "registered_capacity" in fac
                else 0,
            ]
            remapped.append(i)

    remapped = sorted(remapped, key=itemgetter(3, 0, 5, 7))

    with open("data/facility_diff/facilities_v3.csv", "w") as fh:
        csvwriter = csv.writer(fh)
        for line in remapped:
            csvwriter.writerow(line)



from typing import List, Optional

from fastapi import Depends, FastAPI
from pydantic import BaseModel
from sqlalchemy.orm import Session, sessionmaker

from opennem.controllers.stations import (
    create_station,
    get_station,
    get_stations,
)
from opennem.db import db_connect
from opennem.db.models.opennem import Facility, Station
from opennem.schema.opennem import StationSchema, StationSubmission

app = FastAPI()

engine = db_connect()
session = sessionmaker(bind=engine, autocommit=False, autoflush=False,)


def get_database_session():
    try:
        s = session()
        yield s
    finally:
        s.close()


@app.get(
    "/stations", response_model=List[StationSchema],
)
def stations(
    db: Session = Depends(get_database_session),
    name: Optional[str] = None,
    limit: Optional[int] = None,
    page: int = 1,
) -> List[StationSchema]:
    stations = get_stations(db, name=name, limit=limit, page=page)

    return stations


@app.get("/station/{station_id}")
def station(
    session: Session = Depends(get_database_session), station_id: int = None
):
    return get_station(session, station_id)


@app.post("/station")
def station_create(
    session: Session = Depends(get_database_session),
    station: StationSubmission = None,
):
    return True
